#
# AnImAl
#
##
## animage - structure definition as S4 method
##
setClass("animage",
         representation =
         representation(data     = "array",         # data array (only 2D currently)
                        pmode    = "character",     # only "scalar" is valid (appropriate for 2D)
                        focus    = "integer",       # data offsetting
                        outside  = "numeric",       # image value outside data region 
                        storage  = "character",     # storage mode spec (only "integer": is valid)
                        document = "logical",       # TRUE 
                        history  = "call"),         # keep track of operations performed
         prototype =
         list(data     = array(as.integer(0), c(3,3)),
              pmode    = "scalar",
              focus    = as.integer(c(0,0)),
              outside  = as.integer(NA),
              storage  = "integer",
              document = FALSE))

# 
# We want valid objects to get terse C interfaces
#
.Animage.valid <- function(object) {
  #
  # animages represent planar patches
  #
  if(length(object@focus) != 2) {
    stop("illegal animage FOCUS specification")
  }

  #
  # animages pixels must CURRENTLY be scalars
  #
  if(length(dim(object@data)) != 2) {
    stop("illegal animage DIMENSIONS specification")
  }

  #
  # we need to insert the checks for the
  # available modes 
  #
  if(!identical(object@pmode,"scalar")) {
    stop(paste(c("illegal animage PMODE specification", pmode), " "))
  }
}

setValidity("animage", .Animage.valid)

#
# Testing 
# 
is.animage <- function (x) {
  is(x, "animage")
}

is.animageList <- function(x) {
  if(typeof(x) == "list") {
    animageList <- TRUE

    for(i in x) {
      animageList <- animageList && is.animage(i)
    }
    return(animageList)
  } else {
    return(FALSE)
  }
}

is.documentedAnimageList <- function(x) {
  if(typeof(x) == "list") {
    documentedAnimageList <- TRUE

    for(i in x) {
      documentedAnimageList <- (documentedAnimageList && is.animage(i) && i@document)
    }
    return(documentedAnimageList)
  } else {
    return(FALSE)
  }
}


#
# The 'user' constructor function
#
animage <- function(data,
                    pmode,
                    focus    = c(0L,0L),
                    outside  = 0L,
                    storage  = "integer",
                    document = FALSE,
                    history  = call("<undef>")) {

  if(missing(pmode)) {
    if(length(dim(data)) == 2) {
      pmode = "scalar";
    } else {
                                        # This is a stub: "vector" is not valid and
                                        # will be intercepted by the 'valid' check
      pmode = "vector";
                                 
    }
  }

  if(document && (history == call("<undef>")))
    history <- match.call()
  
  #
  # storage is important both for computations and
  # for interfacing C language.
  # Currently AnImAl only supports integer images
  #
  if(storage == "integer") {
    res <-
      new("animage",
          data     = array(as.integer(data), dim(data)),
          pmode    = pmode,
          focus    = as.integer(focus),
          outside  = as.integer(outside),
          storage  = "integer",
          document = document)
  } else if (storage == "real") {
    res <-
      new("animage",
          data     = array(as.real(data), dim(data)),
          pmode    = pmode,
          focus    = as.integer(focus),
          outside  = as.real(outside),
          storage  = "real",
          document = document)
  } else {
    return("unsupported animage STORAGE specification")
  }

  if(document)
    res@history = history

  res
}

#
# Print methods
#
.Animage.print <- function(x) {
  noquote(paste("ai",
                " size: ",
                paste(dim(x@data),
                      collapse = " x "),
                " focus: (",
                paste(x@focus,
                      collapse = ", "),
                "), outside: ",
                paste(x@outside,
                      collapse = " "),
                ", storage: ",
                x@storage,
                ", document: ",
                x@document,
                sep = ""))
                
}

print.animage <- function(x, ...) {
  print(.Animage.print(x, ...))
}

setMethod("show", "animage",
          function(object) {
            print.animage(object)
          })

setMethod("plot", "animage",
          function(x, y, ...) ia.show(x, ...)
          )

hist.animage <- function(x, ...) {
  hist(x@data, ...)
}

#
# Coercion methods: we want to promote simple arrays to animages
# but we also want to use coercion to check arguments for .Call
#
setGeneric("as.animage",
           function(x) {
             standardGeneric("as.animage")
           })

setMethod("as.animage", signature(x = "animage"),
          function(x) {
              return(x)
            })

setMethod("as.animage", signature(x = "pixmapRGB"),
          function(x) {
              return(list(as.animage(getChannels(x, colors=c("red"))),
                          as.animage(getChannels(x, colors=c("green"))),
                          as.animage(getChannels(x, colors=c("blue")))))
            })

#
# All default values are appropriate for this
# simple transformation
#
setMethod("as.animage", signature(x = "array"),
          function(x) {
            if(is.integer(x)) {
              animage(data    = x, storage = "integer")
            } else {
              animage(data    = x, storage = "real")
            }
          })

#
# overloading get element to comply with focus
#
setMethod("[", "animage",
          function(x, i, j, drop) {
            if((is.animage(i) || is.animask(i)) && missing(j)) {
              return(aniGet(x, i))
            } else {
              return(aniGet(x, as.integer(c(i, j))))
            }
          })

#
# overloading set element to comply with focus
#
setReplaceMethod("[", signature(x = "animage"),
                 function(x, i, j, value) {
                   aniSet(x, as.integer(c(i, j)), value)
                   return(x)
                 })


##
## animask - structure definition as S4 method
##
setClass("animask",
         representation =
         representation(x    = "integer", # the coordinates of the mask origin 
                        y    = "integer", #
                        w    = "integer", # and the extents: width
                        h    = "integer"),#                  height
         prototype =
         list(x = as.integer(0), y = as.integer(0), w = as.integer(1), h = as.integer(1)))

# 
# We want valid objects to get terse C interfaces
#
.Animask.valid <- function(object) {

  if(object@w < 0) {
    return("illegal animask W specification")
  }

  if(object@h < 0) {
    return("illegal animask H specification")
  }
  
}

setValidity("animask", .Animask.valid)

#
# Testing 
# 
is.animask <- function (x) {
  is(x, "animask")
}

#
# The 'user' constructor function
#
animask <- function(x, y, w, h) {

  new("animask",
      x = as.integer(x),
      y = as.integer(y),
      w = as.integer(w),
      h = as.integer(h))
      
}

#
# Print methods
#
.Animask.print <- function(x) {
  noquote(paste("am",
                "size: ",
                x@w, " x " , x@h,
                "focus: (", x@x, ",", x@y, ")"))
}

print.animask <- function(x, ...) {
  print(.Animask.print(x, ...))
}

setMethod("show", "animask",
          function(object) {
            print.animask(object)
          })

#
# Coercion methods: we want to promote simple arrays to animages
# but we also want to use coercion to check arguments for .Call
#
setGeneric("as.animask",
           function(x) {
             standardGeneric("as.animask")
           })

#
# We provide coercion for two simple cases: list and vectors
# of 2 (4) elements
#
setMethod("as.animask", signature(x = "integer"),
          function(x) {
            if(length(x) == 2) {
              animask(0L,
                      0L,
                      x[1],
                      x[2])
            } else if (length(x) == 4) {
              animask(x[1],
                      x[2],
                      x[3],
                      x[4])
            } else {
              stop("illegal animask COERCION")
            }
          })

setMethod("as.animask", signature(x = "list"),
          function(x) {
            if(length(x) == 2) {
              animask(0L,
                      0L,
                      x[[1]],
                      x[[2]])
            } else if (length(x) == 4) {
              animask(x[[1]],
                      x[[2]],
                      x[[3]],
                      x[[4]])
            } else {
              stop("illegal animask COERCION")
            }
          })

