# AnImAl
#
#
# R function wrappers for png and tiff IO
#

tiffstat <- function(filename)
  {
    info <- .Call("at_tiffstat", filename, PACKAGE="AnImAl")

    list(SAMPLEFORMAT=info[[1]],
         SAMPLESPERPIXEL=info[[2]],
         BITSPERSAMPLE=info[[3]],
         EXTRASAMPLES=info[[4]],
         IMAGELENGTH=info[[5]],
         IMAGEWIDTH=info[[6]])
  }

readtiff <- function(filename)
{
   .Call("at_readtiff", filename, PACKAGE="AnImAl")
}

readpng <- function(filename)
{
   .Call("at_readpng_rgba", filename, PACKAGE="AnImAl")
}

writetiff <- function(imageChannels, filename, dataFormat = 0, hasTransparency = FALSE)
{
  if(is.animage(imageChannels))
    imageChannels <- c(imageChannels);

  # we must check that all image channels are integer
  # animages or that a float tiff format has been requested
  #
  for (c in imageChannels) 
    if(!is.animage(c) ||
       ((dataFormat != 2) && !(c@storage == "integer")))
      stop("image channels must be integer animages");

  # and that all of them share the same
  # dimensions
  #
  channelsSize <- dim(imageChannels[[1]]@data);

  if(length(imageChannels) > 1)
    for(c in 2:length(imageChannels)) {
      cdim <- dim(imageChannels[[c]]@data)
      
      if((cdim[1] != channelsSize[1]) ||
         (cdim[2] != channelsSize[2]))
        stop("all image channels must have the same size");
    }
  
  # the number of channels must be 1 or 3 (2 or 4, if hasTransparency
  # is true)
  if(hasTransparency) {
     if((length(imageChannels) != 2) &&
        (length(imageChannels) != 4)) {
       stop("illegal number of channels")
     }
   } else if ((length(imageChannels) != 1) &&
              (length(imageChannels) != 3)) {
     stop("illegal number of channels")
   }
  
  .Call("at_writetiff",
        imageChannels,
        hasTransparency,
        filename, dataFormat, PACKAGE="AnImAl")
}

readimage <- function (filename) {

  suffixCandidates <- strsplit(filename, "\\.")
  suffix           <- (suffixCandidates[[1]])[[length(suffixCandidates[[1]])]]
  
  if(suffix == "tif") {
    channelData  <- readtiff(filename);
  } else {
    channelData  <- readpng(filename);
  }
  
  channelData
}

ia.readAnimageChannel <- function (info, channel, document = FALSE) {

  if(typeof(info) == "character") {
    filename <- info
    suffixCandidates <- strsplit(filename, "\\.")
    suffix           <- (suffixCandidates[[1]])[[length(suffixCandidates[[1]])]]
  
    if(suffix == "tif") {
      channelData  <- readtiff(filename);
    } else {
      channelData  <- readpng(filename);
    }
  } else {
    channelData <- info
  }

  if(document) {
    resh          <- match.call()
    resh$info     <- info
    resh$channel  <- channel
    resh$document <- document
  } else {
    resh <- call("<undef>")
  }

  if(typeof(channelData[[channel]]) == "integer") {
    animage(data     = channelData[[channel]],
            document = document,
            history  = resh)
  } else {
    animage(data     = channelData[[channel]],
            document = document,
            storage  = "real",
            history  = resh)
  }

}

#
# NOTE: we should find a nicer way to cope with history. Whenever
# a self documenting image is required, the image is read multiple
# times, once per channel, in order to have a coherent history without
# expanding array data
#
ia.readAnimage <- function (filename, document = FALSE) {
  suffixCandidates <- strsplit(filename, "\\.")
  suffix           <- (suffixCandidates[[1]])[[length(suffixCandidates[[1]])]]
  
  if(suffix == "tif") {
    channels <- tiffstat(filename)[[2]]
  } else {
    channels <- 4;
  }
  
  if(document) {
    res         <- c()

    for(channel in 1:channels)
      res <- c(res, ia.readAnimageChannel(filename, channel, document))
  } else {
    channelData <- readimage(filename)
    res         <- c()

    for(channel in 1:channels)
      res <- c(res, ia.readAnimageChannel(channelData, channel, document))
  }
  res
  
}

#
# Stretches or compresses img dynamic range to match dstRange by linear
# scaling
#
ia.stretchRange <- function(img,
                            dstRange = c(0,255)) {
  data     <- img@data
  minvalue <- min(data)
  maxvalue <- max(data)

  data     <- (((data - minvalue) / (maxvalue - minvalue)) *
               (dstRange[2] - dstRange[1])) + dstRange[1]

  res      <- animage(data     = data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = img@focus)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
}

#
# Similar to ia.stretchRange but it additionally requires a
# srcRange (defaults to c(0,255)): values lower (higher) than the
# minimum (maximum) value are clipped to srcRange and mapped
# accordingly
#
ia.mapRange <- function(img,
                        srcRange  = c(0,255),
                        dstRange  = c(0,255)) {
  data     <- img@data
  minvalue <- srcRange[1]
  maxvalue <- srcRange[2]

  data[data < minvalue] <- minvalue
  data[data > maxvalue] <- maxvalue
  

  data     <- (((data - minvalue) / (maxvalue - minvalue)) *
               (dstRange[2] - dstRange[1])) + dstRange[1]

  res      <- animage(data     = data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = img@focus)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
}

ia.clipRange <- function(img,
                         clipRange  = c(0,255)) {
  data     <- img@data
  minvalue <- clipRange[1]
  maxvalue <- clipRange[2]

  data[data < minvalue] <- minvalue
  data[data > maxvalue] <- maxvalue
  
  res      <- animage(data     = data,
                      pmode    = img@pmode,
                      outside  = img@outside,
                      storage  = img@storage,
                      document = img@document,
                      focus    = img@focus)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }

  res
}

#
# Compute a weighted linear combination of image channels
# (e.g. to obtain a grey level image)
#
ia.averageImageChannels <- function(img,
                                    channels,
                                    weights) {

  if(!is.list(img)) {
    return(img);
  }
   
  W   <- 0L
  if(img[[1]]@storage == "integer") {
    weights <- as.integer(weights)
  }

  document                    <- img[[channels[1]]]@document

  img[[channels[1]]]@document <- FALSE
  
  res <- ia.mult(img[[channels[1]]], weights[channels[1]])
  W   <- W + weights[channels[1]]
  
  img[[channels[1]]]@document <- document

  if(length(channels) > 1) {
    for (i in 2:length(channels)) {
      res <- ia.add(res, ia.mult(img[[channels[i]]], weights[channels[i]]))
      W   <- W + weights[channels[i]]
    }
  }
  
  res <- ia.div(res, W)

  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  
  res
}

