#
# \litcode{Edge detection for a binary image}\index{edge~detection!binary images}
#
# This is an example of how the basic bit level logical
# operators provided by AnImAl can be used to extend its functionalities.
# In this particular case we define a new function with name \texttt{ia\_edge}
# and a single argument \texttt{img}:
#
ia.edge <- function (img) {
  #
  # The first step is to check whether image history must be updated in
  # order to produce a coherent image:
  #
  document <- img@document
  #
  # We momentarily turn off self documentation as we want to consider
  # the current function as a whole and we do not want 
  # its inner workings to be traced by the history mechanism:
  #
  img@document <- FALSE
  #
  # In this particular case we fix the region of interest of the resulting image
  # to that of the function argument. As subsequent
  # operations may enlarge we store the initial specification of the region
  #
  img_mask <- ia.mask(img)
  #
  # A pixel is considered to be an edge pixel
  # if at least one of the pixels to its left, right, top, or 
  # bottom belongs to the image background (value=0):
  # \[
  #     \mat{E}_{i,j} \gets \overline{(((\mat{I}_{i,j-1}   \wedge
  #                                      \mat{I}_{i,j+1})  \wedge
  #                                      \mat{I}_{i+1,j})  \wedge
  #                                      \mat{I}_{i-1,j})} \wedge \mat{I}_{i,j}
  # \]
  # This logical formula can be easily implemented using the bit level
  # logical operations provided by AnImAl:
  #
  res <- ia.get(ia.and(img,
                       ia.not(ia.and(ia.up(img),
                                     ia.and(ia.down(img),
                                            ia.and(ia.left(img),
                                                   ia.right(img)
                                                   ))))),
                img_mask)
  # 
  # If required by the configuration of the argument  image \texttt{img}
  # we update the history of the result image \texttt{res} to 
  # keep it coherent 
  #
  if(document) {
    # In this case, it is necessary to update the tracing flag: 
    img@document <- res@document <- document
    # and to get the new root of the history tree which is given by the current
    # function invocation:
    resh         <- match.call()
    # We then need to expand the node corresponding to \code{img} with the
    # corresponding history 
    resh$img     <- img@history
    # so that complete history information can be stored in the resulting
    # image:
    res@history  <- resh
  }
  # before returning the final result:
  res }
