\name{ia.correlation}

\alias{ia.correlation}

\title{Correlation }

\description{

   Correlate an image with a (image) template

}

\usage{ia.correlation(img,
                      template,
                      dd        =  0L,
                      type      =  "Pearson",
                      normalize =  TRUE,
                      p         =  2L,
                      range     =  1,
                      filter    =  FALSE,
                      mask      =  FALSE)}

\arguments{
    \item{img}{a  real \code{animage}}
    \item{template}{a  real \code{animage}}
    \item{dd}{an integer value 0/1}
    \item{type}{one of "Pearson", "G", "L", or "Lp"}
    \item{normalize}{a logical value}
    \item{p}{an integer representing the metrics exponent}
    \item{range}{a real representing image range}
    \item{filter}{FALSE or a real \code{animage}}
    \item{mask}{FALSE or a real \code{animage}}
}

\value{
 A real \code{animage} whose values are in the range [-100,100].
}

\details{

The function  computes the correlation of the  given \code{animage} with
the provided template.  If \code{normalize  = TRUE} the template and the
corresponding  image  patch are  normalized  to  zero  average and  unit
standard  deviation.   There  are  four different  types  of  similarity
measures  available:   the  standard  Pearson   correlation  coefficient
(\code{type  = "Pearson"}), a  non robust  L1 based  correlation measure
(\code{type = "G"}), a robust L1 based correlation measure (\code{type =
"L"}), and an Lp based  measure (\code{type = "Lp"}). With the exception
of  the  Pearson  case,  setting  \code{normalize =  FALSE}  results  in
meaningful results  (in the  Pearson case, the  unnormalized correlation
coefficient,  a dot  product, is  computed). The  estimator based  on Lp
requires the correct specification of  the range of pixel differences in
order to  transform the Lp  distance value into a  normalized similarity
measure lying in the interval [-1,1].

The computation  can take advantage  of the optional specification  of a
filter image, identifying the image positions where the computation must
take place, and of a template (image) mask, that provides the weight for
each  pixel  contribution. The  latter  can be  used  to  work with  non
rectangular templates and with  templates whose pixel variability varies
significantly.

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}


\references{
  Brunelli, R., Messelodi, S. (1995).
  \emph{Robust Estimation of Correlation with Applications to Computer Vision},
  Pattern Recognition {\bold{28}}:833-841
}

\examples{
#
# get paths to package samples
#
sampleimages <- file.path(system.file(package = "AnImAl"), "sampleimages/")
#
# Read a sample image
#
sampleFace <- as.animage(getChannels(read.pnm(file.path(sampleimages,
"sampleFace_01.pgm"))))
sampleFaceAlt <- ia.mult(sampleFace, 0.5)
eyeMask       <- animask(30, 90, 50, 40)
sampleEye     <- ia.get(sampleFace, eyeMask)
#
cPearson      <- ia.correlation(sampleFace,    sampleEye, type="Pearson")
cPearsonA     <- ia.correlation(sampleFaceAlt, sampleEye, type="Pearson")
cPearsonB     <- ia.correlation(sampleFaceAlt, sampleEye,
normalize=FALSE,  type="Pearson")
#
oldpar        <- par(no.readonly = TRUE)
#
par(mfcol   = c(2,3),
    mar     = c(2,0,0,0),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0),
    cex.lab = 1.5)
#
ia.show(sampleFace)
ia.show(cPearson[[1]],  xlab="Normalized") 
ia.show(sampleFaceAlt)
ia.show(cPearsonA[[1]], xlab="Normalized")
ia.show(sampleFaceAlt)
ia.show(cPearsonB[[1]], xlab="Not normalized")
#
mtext("Pearson correlation and scaling/normalization effects",
       outer = TRUE, side = 3, cex = 1.4, line =1)
#
par(oldpar)
#
cG            <- ia.correlation(sampleFace, sampleEye, type="G")
cL            <- ia.correlation(sampleFace, sampleEye, type="L")
#
par(mfrow   = c(2,3),
    mar     = c(1,0,0,0),
    mgp     = c(1,1,1),
    oma     = c(0,0,3,0))
#
persp(ia.subsample(cPearson[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
persp(ia.subsample(cG[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
persp(ia.subsample(cL[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
mtext("Pearson, G, and L compared",
       outer = TRUE, side = 3, cex = 1.4, line =1)
#
par(mar     = c(4,4,0,1),
    mgp     = c(2,0.5,0.0),
    lwd     = 0.5)
#
hist(cPearson[[1]]@data, xlab = "Pearson", main="", probability = TRUE)
hist(cG[[1]]@data, xlab = "G", ylab = "",  main="", probability = TRUE)
hist(cL[[1]]@data, xlab = "L", ylab = "",  main="", probability = TRUE)
#
cLp1          <- ia.correlation(sampleFace, sampleEye, type="Lp", p=1L,
                                range = 2)
cLp2          <- ia.correlation(sampleFace, sampleEye, type="Lp", p=2L,
                                range = 2)
cLp3          <- ia.correlation(sampleFace, sampleEye, type="Lp", p=3L,
                                range = 2)
#
par(mfrow   = c(2,3),
    mar     = c(1,0,0,0),
    mgp     = c(1,1,1),
    oma     = c(0,1,3,0))
#
persp(ia.subsample(cLp1[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
persp(ia.subsample(cLp2[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
persp(ia.subsample(cLp3[[1]], 2)@data,
      xlab="x", ylab="y", theta = 5, phi = 30,
      shade = 0.5, r = 1,lwd=0.1)
#
par(mar     = c(4,4,0,1),
    mgp     = c(2,0.5,0.0),
    lwd     = 0.5)
#
hist(cLp1[[1]]@data, xlab = "p = 1", main="", probability = TRUE)
hist(cLp2[[1]]@data, xlab = "p = 2", ylab = "",  main="", probability = TRUE)
hist(cLp3[[1]]@data, xlab = "p = 3", ylab = "",  main="", probability = TRUE)
#
mtext("Lp dependency on p",
       outer = TRUE, side = 3, cex = 1.4, line =1)
#
}
\keyword{array}
