#include <arrayTools/Correlation.h>
#include <limits.h>

namespace _ARRAYTOOLS_NAMESPACE
{

  using std::sqrt; 

  //! The following functions are provided to map the correlation values
  //! obtained through L1 estimators to the standard correlation values
  //! This is necessary to avoid providing different thresholds for
  //! each of the correlation types
  //!

  double rd_1(double *px, double *py)
  {
	double	r;

	r = (px[0] - px[1]) / (py[0] - py[1]);

	return(r);
  }
    
  double rd_2(double *px, double *py)
  {
	double r;

	r = py[1] + (px[0] - px[2]) / (rd_1(px, py) - rd_1(px+1,py+1));

	return(r);
  }

  double rd_3(double *px, double *py)
  {
	double r;

	r = rd_1(px+1,py+1) +
	  (px[0] - px[3]) / (rd_2(px, py) - rd_2(px+1,py+1));

	return(r);
  }

  double rd_4(double *px, double *py)
  {
	double r;

	r = rd_2(px+1,py+1) +
	  (px[0] - px[4]) / (rd_3(px, py) - rd_3(px+1,py+1));

	return(r);
  }
    
  double ratint(double x, double *px, double *py)
  {
	return(py[0]+((x-px[0]) /
				  (rd_1(px,py)+((x-px[1]) /
								(rd_2(px,py)-py[0]+((x-px[2]) /	
													(rd_3(px,py)-rd_1(px,py)+((x-px[3])/
																			  (rd_4(px,py)-rd_2(px,py))))))))));
  }

  //! The following are numerically computed tables for mapping
  //! L1 correlation value to the standard ones.
  //!

  double ltx[14] = {   0.0,
					   0.22540460,
					   0.25838141,  
					   0.29289467,
					   0.32918104,
					   0.36754572,
					   0.40839759,
					   0.45227873,
					   0.50000140,
					   0.55278801,
					   0.61270303,
					   0.68377221,
					   0.77639307, 
					   1.00000000
  };

  double lty[14] = { -1.000,
					 -0.200,
					 -0.100,
					 0.000,
					 0.100,
					 0.200,
					 0.300,
					 0.400,
					 0.500,
					 0.600,
					 0.700,
					 0.800,
					 0.900,
					 1.000
  };

  double ltxb[22] = {  0.00000000,
					   0.21614728,
					   0.23676163,
					   0.25773439,
					   0.27936483,
					   0.30197669,
					   0.32595182,
					   0.35177441,
					   0.38010332,
					   0.41189904,
					   0.44868328,
					   0.49314025,
					   0.55085602,
					   0.63767005,
					   0.64964524,
					   0.66266662,
					   0.67696866,
					   0.69288215,
					   0.71089417,
					   0.73177512,
					   0.75686773,
					   1.00000000};

  double ltyb[22] = {-1.000,
					 -0.300,
					 -0.200,
					 -0.100,
					 0.000,
					 0.100,
					 0.200,
					 0.300,
					 0.400,
					 0.500,
					 0.600,
					 0.700,
					 0.800,
					 0.900,
					 0.910,
					 0.920,
					 0.930,
					 0.940,
					 0.950,
					 0.960,
					 0.970,
					 1.000
  };


  //! Given a value and two tables returns the corresponding mapped
  //! value through rational interpolation
  //!

  double remap(double xx, double *px, double *py, int n)
  {
	int		i;
	int		idx;
	double	x;

	x = MAX(px[0], MIN(px[n-1], xx));
   
	i = 0;
   
	while((i < n) && (x > px[i]))
	  i++;

	idx	= MIN(n-1, i+2);

	if(idx > 4)
	  idx 	-= 4;
	else
	  idx	= 0;

	return(ratint(x, px+idx, py+idx));
  }


  //! Given the value and the type of correlation used returns a standard
  //! value
  //!

  double  RemapCorrelationValue(double value, int kind)
  {
	double remappedValue = value;
   
	switch(kind)
      {
	  case 'S':
	  case 'P':
		break;
	  case 'G':
		remappedValue = remap(value, &(ltx[0]), &(lty[0]), 14); 
		break;
	  case 'L':
		remappedValue =  remap(value, &(ltxb[0]), &(ltyb[0]), 22);  
		break;
	  default:
		cerr << "Illegal correlation type!\n";
		exit(-1);
      }
   
	remappedValue = MAX(-1.0, MIN(1.0, remappedValue));
   
	return(remappedValue);
  }


  //! Compute the unnormalized cross-correlation coefficient.
  //! Non rectangular patches can be used, both for the template and
  //! the search region. Local deformation can also be introduced
  //!
  //!

  array<float>& UNCorrelation(array<float>&  II,   //! Image where the template is searched for
							  array<float>&  TT,   //! Template to be located
							  int            dd,   //! Size of neighborhood
							  int            kind, //! can be 'S' 'G' or 'L'
							  array<float>*  F,    //! Filter of image
							  array<float>*  M,    //! Mask of template
							  int           *rx,  //! Return x coordinate of best value
							  int           *ry,  //! Return y coordinate of best value
							  float         *mr   //! Return best value
							  )
  {
	array<float>	*R, T = TT, I = II;
	int		x0, y0;
	int		x1, y1;
	int          c0, c1, c2;
	int          r0, r1, r2;
	int		n, i, j, k, x, y;
	double	rv, tv, dv, rrv, ddv;
	double	MR = -100;
	int		nn[8][2];
	double       sxy, su, sl;
	int          tx1, ty1;
	float        **_T, **_I;


	//! Make sure that the template has zero based indexes
	//!
	T.translate_at(0,0);
   
	//! Pixel offsets in neighborhood
	//!
	nn[0][0]	=  0;
	nn[0][1]	= -1;
	nn[1][0]	=  1;
	nn[1][1]	=  0;
	nn[2][0]	=  0;
	nn[2][1]	=  1;
	nn[3][0]	= -1;
	nn[3][1]	=  0;
	nn[4][0]	=  1;
	nn[4][1]	= -1;
	nn[5][0]	=  1;
	nn[5][1]	=  1;
	nn[6][0]	= -1;
	nn[6][1]	=  1;
	nn[7][0]	= -1;
	nn[7][1]	= -1;
   
   
	//! Get some info
	//!
	c0	= T.cols();
	r0	= T.rows();
	c2	= I.cols();
	r2	= I.rows();
   
	//! Correlation region size
	//!
	c1	= c2 - c0 + 1;
	r1	= r2 - r0 + 1;
   
	//! and limits
	//!
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<float>(y0, y1, x0, x1);

	_T   = T.storage();
	_I   = I.storage();

	//! Count the template pixels
	//!
	n   = 0;
	for(j = 0; j <= ty1; j++)
	  for(i = 0; i <= tx1; i++)
		if((M == NULL) || ((*M)[j][i] > 0))
		  n++;

	//! Compute correlation
	//!
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				//! Reset accumulators
				//!
				sxy = su = sl = 0.0;
		  
				//! loop on template
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if((M == NULL) || ((*M)[j][i] > 0))
						  {
							rv =  _I[j+y][i+x];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _I[y+j+nn[k][0]][x+i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  }
						  }
					  }
				  }
		  
				switch(kind)
				  {
				  case 'S':
					(*R)[y][x]  = sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					(*R)[y][x]  = 100*(1.0 - su / n);
					break;
				  }
		  
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
      }
   
	if(F != NULL)
	  DeleteIfTemporary(*F);

	if(M != NULL)
	  DeleteIfTemporary(*M);

	//! Adjust size 
	//!
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	return((*R).temporary());
  }

  //! Compute the cross-correlation coefficient.
  //! Non rectangular patches can be used, both for the template and
  //! the search region. Local deformation can also be introduced
  //!
  //!

  array<float>& Correlation(array<float>&  II,   //! Image where the template is searched for
							array<float>&  TT,   //! Template to be located
							int            dd,   //! Size of neighborhood
							int            kind, //! can be 'S' 'G' or 'L'
							array<float>*  F,    //! Filter of image
							array<float>*  M,    //! Mask of template
							int           *rx,  //! Return x coordinate of best value
							int           *ry,  //! Return y coordinate of best value
							float         *mr   //! Return best value
							)
  {
	array<float>	*R, T = TT, I = II;
	int		x0, y0;
	int		x1, y1;
	int          c0, c1, c2;
	int          r0, r1, r2;
	int		n, i, j, k, x, y;
	double	rv, tv, dv, rrv, ddv;
	double	dt, di, MR = -100;
	int		nn[8][2];
	double       sx, sy, sxy, sx2, sy2, su, sl;
	int          tx1, ty1;
	array<float>        *P;
	float       **_T, **_I, **_P;


	//! Make sure that the template has zero based indexes
	//!
	T.translate_at(0,0);
	if(M)
	  (*M).translate_at(0,0);
   
	//! Pixel offsets in neighborhood
	//!
	nn[0][0]	=  0;
	nn[0][1]	= -1;
	nn[1][0]	=  1;
	nn[1][1]	=  0;
	nn[2][0]	=  0;
	nn[2][1]	=  1;
	nn[3][0]	= -1;
	nn[3][1]	=  0;
	nn[4][0]	=  1;
	nn[4][1]	= -1;
	nn[5][0]	=  1;
	nn[5][1]	=  1;
	nn[6][0]	= -1;
	nn[6][1]	=  1;
	nn[7][0]	= -1;
	nn[7][1]	= -1;
   
   
	//! Get some info
	//!
	c0	= T.cols();
	r0	= T.rows();
	c2	= I.cols();
	r2	= I.rows();
   
	//! Correlation region size
	//!
	c1	= c2 - c0 + 1;
	r1	= r2 - r0 + 1;
   
	//! and limits
	//!
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<float>(y0, y1, x0, x1);
	P    = new array<float>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();
   
	//! Compute normalization factor for template image
	//!
	sx2 = 0;
	sx  = 0;
	n   = 0;
   
	for(j = 0; j <= ty1; j++)
      {
		for(i = 0; i <= tx1; i++)
		  {
			if((M == NULL) || ((*M)[j][i] > 0))
			  {
				sx   += _T[j][i];
				sx2  += _T[j][i] * _T[j][i];
				n++;
			  }
		  }
      }
	sx2	/= n;
	sx   /= n;
   
	//! Deviation of template
	//!
	dt  	= sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
	//! Normalize template to zero average and unit sigma
	for(j = 0; j <= ty1; j++)
	  for(i = 0; i <= tx1; i++)
		_T[j][i] = (_T[j][i] - sx) / dt;
   
   
	//! Compute correlation
	//!
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				//! Copy image patch into patch area
				//!
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = _I[y+j][x+i];
		  
				//! Reset accumulators
				//!
				sxy = 0;
				sy  = 0;
				sy2 = 0;
				su  = 0;
				sl  = 0;
		  
				//! Normalize image patch
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if((M == NULL) || ((*M)[j][i] > 0))
						  {
							rv  =  _I[y+j][x+i];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }
				  }
				sy  /= n;
				sy2 /= n;
				di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
		  
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = (_P[j][i] - sy) / di;
		  
		  
				//! loop on template
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if((M == NULL) || ((*M)[j][i] > 0))
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  }
						  }
					  }
				  }
		  
				switch(kind)
				  {
				  case 'S':
					sxy	    /= n;
					sxy          = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					(*R)[y][x]  = 100*(1.0 - su / n);
					break;
				  }
		  
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
      }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);

	if(M != NULL)
	  DeleteIfTemporary(*M);

	//! Adjust size 
	//!
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	*mr = RemapCorrelationValue(*mr, kind);

	return((*R).temporary());
  }



  array<float>& HierarchicalCorrelation(array<float>* image_pyramid,
										array<float>* template_pyramid,
										int    levels,
										int*   neighborhood,
										int    kind, //! can be 'S' 'G' or 'L'
										double fraction_to_keep,
										int    *rx, //! Return x coordinate of best value
										int    *ry, //! Return y coordinate of best value
										double *mr  //! Return best value
										)
  {
	array<float>    *correlation_map;
	array<float>    *filter;
	array<float>    *R;
	int      best_value_x, best_value_y;
	int      n;
	int      i, l;
	int      x, y;
	int      x0, x1;
	int      y0, y1;
	float   best_value;
	float   dont_keep;
	float   minimum_value_to_keep;
	vector<float>   histogram(101);
	float   **_f;
   
	correlation_map = new array<float>[levels];
	filter          = new array<float>[levels];
   
	filter[0] = (* (new array<float>(image_pyramid[0].first_row_index(),
									 image_pyramid[0].last_row_index(),
									 image_pyramid[0].first_col_index(),
									 image_pyramid[0].last_col_index()))).temporary();
   
	filter[0] = 1.0;
	_f        = filter[0].storage();
   
	for(l = 0; l < levels; l++)
      {
		float **_C;

	
		//! Compute correlation map at the given level 
		//! focussing computation on the sites which
		//! were found to be promising at the 
		//! previous level
		//!
		correlation_map[l] = Correlation(image_pyramid[l],
										 template_pyramid[l],
										 neighborhood[l],
										 kind,
										 &(filter[l]),
										 NULL,
										 &best_value_x, &best_value_y, &best_value);
	
		_C = correlation_map[l].storage();
	
		//! Reset histogram
		//!
		histogram = histogram * 0;
	
		//! Scan correlation map to compute the distribution
		//! of the correlation values
		//!
		x0 = correlation_map[l].first_col_index();
		x1 = correlation_map[l].last_col_index();
		y0 = correlation_map[l].first_row_index();
		y1 = correlation_map[l].last_row_index();
	
	
		//! Normalize correlation values to the interval [0,100]
		//!
		{
		  float  max_c_value = FLT_MIN, min_c_value  = FLT_MAX;
		  int    init = TRUE;
	   
		  n = 0;
		  for(y = y0; y <= y1; y++)
			for(x = x0; x <= x1; x++)
			  if(filter[l][y][x] > 0.0)
				{
				  if(init == TRUE)
					{
					  init = FALSE;
					  max_c_value = min_c_value = correlation_map[l][y][x];
					}
				  else
					{
					  max_c_value = MAX(max_c_value, correlation_map[l][y][x]);
					  min_c_value = MIN(min_c_value, correlation_map[l][y][x]);
					}
				  n++;
				}
	   
		  if(n > 0)
			{
			  for(y = y0; y <= y1; y++)
				for(x = x0; x <= x1; x++)
				  if(filter[l][y][x] > 0.0)
					{
					  correlation_map[l][y][x] = 
						((correlation_map[l][y][x]-min_c_value) /
						 (max_c_value - min_c_value))*100.0;
					}
			}
		}
	
		for(y = y0; y <= y1; y++)
		  for(x = x0; x <= x1; x++)
			if(filter[l][y][x] > 0.0)
			  {
				histogram[int(correlation_map[l][y][x])]++;
				n++;
			  }
	
		//! Determine the minimum correlation value for
		//! which we want to compute correlation at
		//! the next level
		//!
		dont_keep = n*(1-fraction_to_keep);
		n         = 0;
		i         = 0;
		while((n <= dont_keep) && (i <= 100))
		  n += int(histogram[i++]);
	
		//! minimum_value_to_keep = MAX(0,(i-2));
		minimum_value_to_keep = MAX(0,(i-1));
	
		//! Compute the filter to be used at the next level
		//!
		if(l+1 < levels)
		  {
			x0 = image_pyramid[l+1].first_col_index();
			x1 = image_pyramid[l+1].last_col_index();
			y0 = image_pyramid[l+1].first_row_index();
			y1 = image_pyramid[l+1].last_row_index();
	     
			filter[l+1] = correlation_map[l] > minimum_value_to_keep;
			filter[l+1] = Grow(filter[l+1], 1);
			filter[l+1] = Zoom(filter[l+1], 2, 2);
	     
			//! Make sure that dimensions match
			//!
			filter[l+1] = filter[l+1].subimage(y0, y1, x0, x1);
		  }
		else
		  {
			*rx = best_value_x;
			*ry = best_value_y;
			*mr = best_value;
		  }
      }
   
	R = new array<float>(correlation_map[levels-1]);
   
	delete [] correlation_map;
	delete [] filter;
   
	*mr = RemapCorrelationValue(*mr, kind);

	return((*R).temporary());
  }

  array<float>& HierarchicalCorrelation(array<float>* image_pyramid,
										array<float>* template_pyramid,
										array<float>* mask_pyramid,
										int    levels,
										int*   neighborhood,
										int    kind, //! can be 'S' 'G' or 'L'
										double fraction_to_keep,
										int    *rx, //! Return x coordinate of best value
										int    *ry, //! Return y coordinate of best value
										double *mr,  //! Return best value
										vector<double> *bvs,
										int    unusedLevels
										)
  {
	array<float>    *correlation_map;
	array<float>    *filter;
	array<float>    *R;
	int      best_value_x, best_value_y;
	int      n;
	int      i, l;
	int      x, y;
	int      x0, x1;
	int      y0, y1;
	float   best_value;
	float   dont_keep;
	float   minimum_value_to_keep;
	vector<float>   histogram(101);
	float   **_f;
   
	correlation_map = new array<float>[levels];
	filter          = new array<float>[levels];
   
	filter[0] = (* (new array<float>(image_pyramid[0].first_row_index(),
									 image_pyramid[0].last_row_index(),
									 image_pyramid[0].first_col_index(),
									 image_pyramid[0].last_col_index()))).temporary();
   
	filter[0] = 1.0;
	_f        = filter[0].storage();
   
	for(l = 0; l < levels - unusedLevels; l++)
      {
		float **_C;
	
		//! Compute correlation map at the given level 
		//! focussing computation on the sites which
		//! were found to be promising at the 
		//! previous level
		//!
		{
		  array<float> R, T, B;
	   
		  R = image_pyramid[l];
		  T = template_pyramid[l];
		  B = mask_pyramid[l];

		  correlation_map[l] = Correlation(R,
										   T,
										   neighborhood[l],
										   kind,
										   &(filter[l]),
										   &B,
										   &best_value_x, &best_value_y, &best_value);

		  if(bvs)
			(*bvs)[l] = best_value;
		}
	
		_C = correlation_map[l].storage();
	
		//! Reset histogram
		//!
		histogram =  histogram * 0;
	
		//! Scan correlation map to compute the distribution
		//! of the correlation values
		//!
		x0 = correlation_map[l].first_col_index();
		x1 = correlation_map[l].last_col_index();
		y0 = correlation_map[l].first_row_index();
		y1 = correlation_map[l].last_row_index();
	
	
		//! Normalize correlation values to the interval [0,100]
		//!
		{
		  float  max_c_value = FLT_MIN, min_c_value  = FLT_MAX;
		  int    init = TRUE;
	   
		  n = 0;
		  for(y = y0; y <= y1; y++)
			for(x = x0; x <= x1; x++)
			  if(filter[l][y][x] > 0.0)
				{
				  if(init == TRUE)
					{
					  init = FALSE;
					  max_c_value = min_c_value = correlation_map[l][y][x];
					}
				  else
					{
					  max_c_value = MAX(max_c_value, correlation_map[l][y][x]);
					  min_c_value = MIN(min_c_value, correlation_map[l][y][x]);
					}
				  n++;
				}
	   
		  if(n > 0)
			{
			  for(y = y0; y <= y1; y++)
				for(x = x0; x <= x1; x++)
				  if(filter[l][y][x] > 0.0)
					{
					  correlation_map[l][y][x] = 
						((correlation_map[l][y][x]-min_c_value) /
						 (max_c_value - min_c_value))*100.0;
					}
			}
		}
	
		for(y = y0; y <= y1; y++)
		  for(x = x0; x <= x1; x++)
			if(filter[l][y][x] > 0.0)
			  {
				histogram[int(correlation_map[l][y][x])]++;
				n++;
			  }
	
		//! Determine the minimum correlation value for
		//! which we want to compute correlation at
		//! the next level
		//!
		dont_keep = n*(1-fraction_to_keep);
		n         = 0;
		i         = 0;
		while((n <= dont_keep) && (i <= 100))
		  n += int(histogram[i++]);
	
		minimum_value_to_keep = MAX(0,(i-1));
	
		//! Compute the filter to be used at the next level
		//!
		if(l+1 < levels - unusedLevels)
		  {
			x0 = image_pyramid[l+1].first_col_index();
			x1 = image_pyramid[l+1].last_col_index();
			y0 = image_pyramid[l+1].first_row_index();
			y1 = image_pyramid[l+1].last_row_index();
	     
			filter[l+1] = correlation_map[l] > minimum_value_to_keep;
			filter[l+1] = Grow(filter[l+1], 1);
			filter[l+1] = Zoom(filter[l+1], 2, 2);
	     
			//! Make sure that dimensions match
			//!
			filter[l+1] = filter[l+1].subimage(y0, y1, x0, x1);
		  }
		else
		  {
			*rx = best_value_x;
			*ry = best_value_y;

			//! and scale appropriately
			//!
			for(int i = 0; i < unusedLevels; i++)
			  {
				*rx *= 2;
				*ry *= 2;
			  }
	     
			*mr = best_value;
		  }
      }
   
	R = new array<float>(correlation_map[levels-1]);
   
	delete [] correlation_map;
	delete [] filter;
   
	*mr = RemapCorrelationValue(*mr, kind);

	return((*R).temporary());
  }


  array<double>& HierarchicalCorrelation(array<double>* image_pyramid,
										 array<double>* template_pyramid,
										 int    levels,
										 int*   neighborhood,
										 int    kind, //! can be 'S' 'G' or 'L'
										 double fraction_to_keep,
										 int    *rx, //! Return x coordinate of best value
										 int    *ry, //! Return y coordinate of best value
										 double *mr  //! Return best value
										 )
  {
	array<double>    *correlation_map;
	array<double>    *filter;
	array<double>    *R;
	int      best_value_x, best_value_y;
	int      n;
	int      i, l;
	int      x, y;
	int      x0, x1;
	int      y0, y1;
	double   best_value;
	double   dont_keep;
	double   minimum_value_to_keep;
	vector<double>   histogram(101);
	double   **_f;
   
	correlation_map = new array<double>[levels];
	filter          = new array<double>[levels];
   
	filter[0] = (* (new array<double>(image_pyramid[0].first_row_index(),
									  image_pyramid[0].last_row_index(),
									  image_pyramid[0].first_col_index(),
									  image_pyramid[0].last_col_index()))).temporary();
   
	filter[0] = 1.0;
	_f        = filter[0].storage();
   
	for(l = 0; l < levels; l++)
      {
		double **_C;
	
		//! Compute correlation map at the given level 
		//! focussing computation on the sites which
		//! were found to be promising at the 
		//! previous level
		//!
		correlation_map[l] = Correlation(image_pyramid[l],
										 template_pyramid[l],
										 neighborhood[l],
										 kind,
										 &(filter[l]),
										 NULL,
										 &best_value_x, &best_value_y, &best_value);
	
		_C = correlation_map[l].storage();
	
		//! Reset histogram
		//!
		histogram = histogram * 0;
	
		//! Scan correlation map to compute the distribution
		//! of the correlation values
		//!
		x0 = correlation_map[l].first_col_index();
		x1 = correlation_map[l].last_col_index();
		y0 = correlation_map[l].first_row_index();
		y1 = correlation_map[l].last_row_index();
	
	
		//! Normalize correlation values to the interval [0,100]
		//!
		{
		  double  max_c_value = DBL_MAX, min_c_value = DBL_MIN;
		  int    init = TRUE;
	   
		  n = 0;
		  for(y = y0; y <= y1; y++)
			for(x = x0; x <= x1; x++)
			  if(filter[l][y][x] > 0.0)
				{
				  if(init == TRUE)
					{
					  init = FALSE;
					  max_c_value = min_c_value = correlation_map[l][y][x];
					}
				  else
					{
					  max_c_value = MAX(max_c_value, correlation_map[l][y][x]);
					  min_c_value = MIN(min_c_value, correlation_map[l][y][x]);
					}
				  n++;
				}
	   
		  if(n > 0)
			{
			  for(y = y0; y <= y1; y++)
				for(x = x0; x <= x1; x++)
				  if(filter[l][y][x] > 0.0)
					{
					  correlation_map[l][y][x] = 
						((correlation_map[l][y][x]-min_c_value) /
						 (max_c_value - min_c_value))*100.0;
					}
			}
		}
	
		for(y = y0; y <= y1; y++)
		  for(x = x0; x <= x1; x++)
			if(filter[l][y][x] > 0.0)
			  {
				histogram[int(correlation_map[l][y][x])]++;
				n++;
			  }
	
		//! Determine the minimum correlation value for
		//! which we want to compute correlation at
		//! the next level
		//!
		dont_keep = n*(1-fraction_to_keep);
		n         = 0;
		i         = 0;
		while((n <= dont_keep) && (i <= 100))
		  n += int(histogram[i++]);
	
		minimum_value_to_keep = MAX(0,(i-1));
	
		//! Compute the filter to be used at the next level
		//!
		if(l+1 < levels)
		  {
			x0 = image_pyramid[l+1].first_col_index();
			x1 = image_pyramid[l+1].last_col_index();
			y0 = image_pyramid[l+1].first_row_index();
			y1 = image_pyramid[l+1].last_row_index();
	     
			filter[l+1] = correlation_map[l] > minimum_value_to_keep;
			filter[l+1] = Grow(filter[l+1], 1);
			filter[l+1] = Zoom(filter[l+1], 2, 2);
	     
			//! Make sure that dimensions match
			//!
			filter[l+1] = filter[l+1].subimage(y0, y1, x0, x1);
		  }
		else
		  {
			*rx = best_value_x;
			*ry = best_value_y;
			*mr = best_value;
	     
		  }
      }
   
	R = new array<double>(correlation_map[levels-1]);
   
	delete [] correlation_map;
	delete [] filter;
   
	*mr = RemapCorrelationValue(*mr, kind);

	return((*R).temporary());
  }


  //! Compute the cross-correlation coefficient.
  //! Non rectangular patches can be used, both for the template and
  //! the search region. Local deformation can also be introduced
  //!
  //!

  array<double>& Correlation(array<double>&  II,   //! Image where the template is searched for
							 array<double>&  TT,   //! Template to be located
							 int            dd,   //! Size of neighborhood
							 int            kind, //! can be 'S' 'G' or 'L'
							 array<double>*  F,    //! Filter of image
							 array<double>*  M,    //! Mask of template
							 int           *rx,  //! Return x coordinate of best value
							 int           *ry,  //! Return y coordinate of best value
							 double         *mr   //! Return best value
							 )
  {
	array<double>	*R, T = TT, I = II;
	int		x0, y0;
	int		x1, y1;
	int          c0, c1, c2;
	int          r0, r1, r2;
	int		n, i, j, k, x, y;
	double	rv, tv, dv, rrv, ddv;
	double	dt, di, MR = -100;
	int		nn[8][2];
	double       sx, sy, sxy, sx2, sy2, su, sl;
	int          tx1, ty1;
	array<double>        *P;
	double       **_T, **_I, **_P;


	//! Make sure that the template has zero based indexes
	//!
	T.translate_at(0,0);
   
	//! Pixel offsets in neighborhood
	//!
	nn[0][0]	=  0;
	nn[0][1]	= -1;
	nn[1][0]	=  1;
	nn[1][1]	=  0;
	nn[2][0]	=  0;
	nn[2][1]	=  1;
	nn[3][0]	= -1;
	nn[3][1]	=  0;
	nn[4][0]	=  1;
	nn[4][1]	= -1;
	nn[5][0]	=  1;
	nn[5][1]	=  1;
	nn[6][0]	= -1;
	nn[6][1]	=  1;
	nn[7][0]	= -1;
	nn[7][1]	= -1;
   
   
	//! Get some info
	//!
	c0	= T.cols();
	r0	= T.rows();
	c2	= I.cols();
	r2	= I.rows();
   
	//! Correlation region size
	//!
	c1	= c2 - c0 + 1;
	r1	= r2 - r0 + 1;
   
	//! and limits
	//!
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<double>(y0, y1, x0, x1);
	P    = new array<double>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();
   
	//! Compute normalization factor for template image
	//!
	sx2 = 0;
	sx  = 0;
	n   = 0;
   
	for(j = 0; j <= ty1; j++)
      {
		for(i = 0; i <= tx1; i++)
		  {
			if((M == NULL) || ((*M)[j][i] > 0))
			  {
				sx   += _T[j][i];
				sx2  += _T[j][i] * _T[j][i];
				n++;
			  }
		  }
      }
	sx2	/= n;
	sx   /= n;
   
	//! Deviation of template
	//!
	dt  	= sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
	//! Normalize template to zero average and unit sigma
	for(j = 0; j <= ty1; j++)
	  for(i = 0; i <= tx1; i++)
		_T[j][i] = (_T[j][i] - sx) / dt;
   
   
	//! Compute correlation
	//!
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				//! Copy image patch into patch area
				//!
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = _I[y+j][x+i];
		  
				//! Reset accumulators
				//!
				sxy = 0;
				sy  = 0;
				sy2 = 0;
				su  = 0;
				sl  = 0;
		  
				//! Normalize image patch
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if((M == NULL) || ((*M)[j][i] > 0))
						  {
							rv  =  _I[y+j][x+i];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }
				  }
				sy  /= n;
				sy2 /= n;
				di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
		  
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = (_P[j][i] - sy) / di;
		  
		  
				//! loop on template
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if((M == NULL) || ((*M)[j][i] > 0))
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  }
						  }
					  }
				  }
		  
				switch(kind)
				  {
				  case 'S':
					sxy	    /= n;
					sxy          = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					(*R)[y][x]  = 100*(1.0 - su / n);
					break;
				  }
		  
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
      }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);

	if(M != NULL)
	  DeleteIfTemporary(*M);

	//! Adjust size 
	//!
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);

	*mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }




  //!//!//!//!//!//!//!//!//!//!//!//!/

  //! This is a slightly modified version of the standard Correlation 
  //! function: the mask is used as a set of weights. Please note that
  //! the sum of the weights must be normalized to 1  by the function
  //! providing the filter
  //!
  array<float>& CorrelationW(array<float>&  II,   //! Image where the template is searched for
							 array<float>&  TT,   //! Template to be located
							 int            dd,   //! Size of neighborhood
							 int            kind, //! can be 'S' 'G' or 'L'
							 array<float>*  F,    //! Filter of image
							 array<float>*  M,    //! Mask of template
							 int           *rx,  //! Return x coordinate of best value
							 int           *ry,  //! Return y coordinate of best value
							 float         *mr   //! Return best value
							 )
  {
	array<float>	*R, T = TT, I = II;
	int		x0, y0;
	int		x1, y1;
	int          c0, c1, c2;
	int          r0, r1, r2;
	int		n, i, j, k, x, y;
	double	rv, tv, dv, rrv, ddv;
	double	dt, di, MR = -100;
	int		nn[8][2];
	double       sx, sy, sxy, sx2, sy2, su, sl;
	int          tx1, ty1;
	array<float>        *P;
	float       **_T, **_I, **_P, **_M = NULL;


	//! Make sure that the template has zero based indexes
	//!
	T.translate_at(0,0);
	if(M)
	  (*M).translate_at(0,0);
   
	//! Pixel offsets in neighborhood
	//!
	nn[0][0]	=  0;
	nn[0][1]	= -1;
	nn[1][0]	=  1;
	nn[1][1]	=  0;
	nn[2][0]	=  0;
	nn[2][1]	=  1;
	nn[3][0]	= -1;
	nn[3][1]	=  0;
	nn[4][0]	=  1;
	nn[4][1]	= -1;
	nn[5][0]	=  1;
	nn[5][1]	=  1;
	nn[6][0]	= -1;
	nn[6][1]	=  1;
	nn[7][0]	= -1;
	nn[7][1]	= -1;
   
   
	//! Get some info
	//!
	c0	= T.cols();
	r0	= T.rows();
	c2	= I.cols();
	r2	= I.rows();
   
	//! Correlation region size
	//!
	c1	= c2 - c0 + 1;
	r1	= r2 - r0 + 1;
   
	//! and limits
	//!
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<float>(y0, y1, x0, x1);
	P    = new array<float>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();

	if(M)
	  _M   = (*M).storage();

   
	//! Compute normalization factor for template image
	//!
	sx2 = 0;
	sx  = 0;
	n   = 0;

	if(M == NULL)
      {
		for(j = 0; j <= ty1; j++)
		  {
			for(i = 0; i <= tx1; i++)
			  {
				sx   += _T[j][i];
				sx2  += _T[j][i] * _T[j][i];
				n++;
			  }
		  }

		sx2	/= n;
		sx   /= n;
      }
	else
      {
		for(j = 0; j <= ty1; j++)
		  {
			for(i = 0; i <= tx1; i++)
			  {
				if(_M[j][i] > 0)
				  {
					sx   += _M[j][i] * _T[j][i];
					sx2  += _M[j][i] * _T[j][i] * _T[j][i];
					n++;
				  }
			  }
		  }
      }
   
	//! Deviation of template
	//!
	dt  	= sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
	//! Normalize template to zero average and unit sigma
	for(j = 0; j <= ty1; j++)
	  for(i = 0; i <= tx1; i++)
		_T[j][i] = (_T[j][i] - sx) / dt;
   
   
	//! Compute correlation
	//!
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			int   nk = 2*dd+tx1+1;
			float w;
	     
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				//! Copy image patch into patch area
				//!
				for(j = -dd; j <= dd+ty1; j++)
				  memcpy(&(_P[j][-dd]), &(_I[y+j][x-dd]), sizeof(float)*nk);

				//! Reset accumulators
				//!
				sxy = 0;
				sy  = 0;
				sy2 = 0;
				su  = 0;
				sl  = 0;
		  
				//! Normalize image patch
				//!
				if(M == NULL)
				  {
					for(j =  0; j <= ty1; j++)
					  {
						for(i = 0; i <= tx1; i++)
						  {
							rv  =  _I[y+j][x+i];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }
					sy  /= n;
					sy2 /= n;
				  }
				else
				  {
					for(j =  0; j <= ty1; j++)
					  {
						for(i = 0; i <= tx1; i++)
						  {
							rv  =  _I[y+j][x+i];
							sy  += _M[j][i] * rv;
							sy2 += _M[j][i] * rv * rv;
						  }
					  }
				  }
		  
				di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
		  
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = (_P[j][i] - sy) / di;
		  
		  
				//! loop on template
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if(M == NULL)
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  }
						  } else if ((w = _M[j][i]) > 0) {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }
				 
							switch(kind)
							  {
							  case 'S':
								sxy += w * tv * rv;
								break;
							  case 'G':
								su  += w * ABS(tv - rv);
								sl  += w * (ABS(tv) + ABS(rv));
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += w * ABS(tv - rv) / sl;
								break;
							  }			       
						  }
			    
					  }
				  }

				switch(kind)
				  {
				  case 'S':
					sxy         = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					(*R)[y][x]  = 100*(1.0 - su);
					break;
				  }
		  
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
      }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);
   
	if(M != NULL)
	  DeleteIfTemporary(*M);
   
	//! Adjust size 
	//!
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	*mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }



  array<float>& HierarchicalCorrelationW(array<float>* image_pyramid,
										 array<float>* template_pyramid,
										 array<float>* mask_pyramid,
										 int    levels,
										 int*   neighborhood,
										 int    kind, //! can be 'S' 'G' or 'L'
										 double fraction_to_keep,
										 int    *rx, //! Return x coordinate of best value
										 int    *ry,	   //! Return y coordinate of best value
										 double *mr,		   //! Return best value
										 vector<double> *bvs,
										 int    unusedLevels
										 )
  {
	array<float>    *correlation_map;
	array<float>    *filter;
	array<float>    *R;
	int      best_value_x, best_value_y;
	int      n;
	int      i, l;
	int      x, y;
	int      x0, x1;
	int      y0, y1;
	float   best_value;
	float   dont_keep;
	float   minimum_value_to_keep;
	vector<float>   histogram(101);
	float   **_f;
   
	correlation_map = new array<float>[levels];
	filter          = new array<float>[levels];
   
	filter[0] = (* (new array<float>(image_pyramid[0].first_row_index(),
									 image_pyramid[0].last_row_index(),
									 image_pyramid[0].first_col_index(),
									 image_pyramid[0].last_col_index()))).temporary();
   
	filter[0] = 1.0;
	_f        = filter[0].storage();
   
	for(l = 0; l < levels - unusedLevels; l++)
      {
		float **_C;
	
		//! Compute correlation map at the given level 
		//! focussing computation on the sites which
		//! were found to be promising at the 
		//! previous level
		//!
		{
		  array<float> R, T, B;
	   
		  R = image_pyramid[l];
		  T = template_pyramid[l];
		  B = mask_pyramid[l];
	   
		  correlation_map[l] = CorrelationW(R,
											T,
											neighborhood[l],
											kind,
											&(filter[l]),
											&B,
											&best_value_x, &best_value_y, &best_value);

		  if(bvs)
			{
			  (*bvs)[l] = best_value;
			}
	   
		}
	
		_C = correlation_map[l].storage();
	
		//! Reset histogram
		//!
		histogram = histogram * 0;
	
		//! Scan correlation map to compute the distribution
		//! of the correlation values
		//!
		x0 = correlation_map[l].first_col_index();
		x1 = correlation_map[l].last_col_index();
		y0 = correlation_map[l].first_row_index();
		y1 = correlation_map[l].last_row_index();
	
		//! Normalize correlation values to the interval [0,100]
		//!
		{
		  float  max_c_value = FLT_MIN, min_c_value  = FLT_MAX;
		  int    init = TRUE;
	   
		  n = 0;
		  for(y = y0; y <= y1; y++)
			for(x = x0; x <= x1; x++)
			  if(filter[l][y][x] > 0.0)
				{
				  if(init == TRUE)
					{
					  init = FALSE;
					  max_c_value = min_c_value = correlation_map[l][y][x];
					}
				  else
					{
					  max_c_value = MAX(max_c_value, correlation_map[l][y][x]);
					  min_c_value = MIN(min_c_value, correlation_map[l][y][x]);
					}
				  n++;
				}
	   
		  if(n > 0)
			{
			  for(y = y0; y <= y1; y++)
				for(x = x0; x <= x1; x++)
				  if(filter[l][y][x] > 0.0)
					{
					  correlation_map[l][y][x] = 
						((correlation_map[l][y][x]-min_c_value) /
						 (max_c_value - min_c_value + 1))*100.0;
					}
			}
		}

	
		n = 0;
		for(y = y0; y <= y1; y++)
		  for(x = x0; x <= x1; x++)
			if(filter[l][y][x] > 0.0)
			  {
				histogram[int(correlation_map[l][y][x])]++;
				n++;
			  }
	
		//! Determine the minimum correlation value for
		//! which we want to compute correlation at
		//! the next level
		//!
		dont_keep = n*(1-fraction_to_keep);
		n         = 0;
		i         = 0;
		while((n <= dont_keep) && (i <= 100))
		  n += int(histogram[i++]);
	
		minimum_value_to_keep = MAX(0,(i-1));
	
		//! Compute the filter to be used at the next level
		//!
		if(l+1 < levels - unusedLevels)
		  {
			x0 = image_pyramid[l+1].first_col_index();
			x1 = image_pyramid[l+1].last_col_index();
			y0 = image_pyramid[l+1].first_row_index();
			y1 = image_pyramid[l+1].last_row_index();
	     
			filter[l+1] = correlation_map[l] > minimum_value_to_keep;
			filter[l+1] = Grow(filter[l+1], 1);
			filter[l+1] = Zoom(filter[l+1], 2, 2);
	     
			//! Make sure that dimensions match
			//!
			filter[l+1] = filter[l+1].subimage(y0, y1, x0, x1);
		  }
		else
		  {
			*rx = best_value_x;
			*ry = best_value_y;


			//! and scale appropriately
			//!
			for(int i = 0; i < unusedLevels; i++)
			  {
				*rx *= 2;
				*ry *= 2;
			  }
	     
			*mr = best_value;
	     
		  }
      }
   
	R = new array<float>(correlation_map[levels-1]);
   
	delete [] correlation_map;
	delete [] filter;

	//! *mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }

  double Correlation(array<double>& I,   // Image where the template is searched for
					 array<double>& T)   // Template to be located
  {
	int		xx, x0, y0, ox = 0;
	int		yy, x1, y1, oy = 0;
	int          c0, r0;
	int		n, i, j, x, y;
	double	rv;
	double	dt, di;
	double       M[2] = {0,0};
	double       sx = 0.0, sy = 0.0, sxy = 0.0, sx2 = 0.0, sy2 = 0.0;
	int          tx0, tx1, ty0, ty1;
	double        mr = 0.0, cvalue = 0.0, nn;
	double        **_T, **_I;

	// Get some info
	//
	c0	= T.cols();
	r0	= T.rows();
   
	// and limits
	//
	x0   = I.first_col_index();
	y0   = I.first_row_index();
	x1   = I.last_col_index() - c0 + 1;
	y1   = I.last_row_index() - r0 + 1;
	tx0  = T.first_col_index();
	ty0  = T.first_row_index();
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	_T   = T.storage();
	_I   = I.storage();
   
	// Compute normalization factor for template image
	//
	sx2 = 0;
	sx  = 0;
	n   = 0;
   
	for(j = ty0; j <= ty1; j++)
      {
		for(i = tx0; i <= tx1; i++)
		  {
			sx   += _T[j][i];
			sx2  += _T[j][i] * _T[j][i];
			n++;
		  }
      }
	sx2	/= n;
	sx   /= n;
	nn    = double(n) * double(n);
   
	// Deviation of template
	//
	dt  	= sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
	// Compute correlation
	//
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			double adjustment;

			sxy = 0;

			if(x == x0)
			  {
				ox  = x0;

				if(y == y0)
				  {		
					int yy1 = y + r0 - 1;
					int xx1 = x + c0 - 1;
       
					sy  = sy2 = 0;
		       
					for(yy = y; yy <= yy1; yy++)
					  {
						for(xx = x; xx <= xx1; xx++)
						  {
							rv  =  _I[yy][xx];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }

					di  = (sy2/n - (sy * sy / nn));
					di  = sqrt(MAX(0.0, di));

					oy   = y0;
					M[0] = sy;
					M[1] = sy2;
				  }
				else
				  {
					int xx1 = x + c0 - 1;
					int ny = oy + r0;
		       
					sy  = M[0];
					sy2 = M[1];
		       
					for(xx = x; xx <= xx1; xx++)
					  {
						rv  =  _I[oy][xx];
						sy  -= rv;
						sy2 -= rv * rv;

						rv  =  _I[ny][xx];
						sy  += rv;
						sy2 += rv * rv;
					  }
		       
					oy++;
					M[0] = sy;
					M[1] = sy2;
					di   = sy2/n - (sy * sy) / nn;
					di   = sqrt(MAX(0.0, di));
				  }
			  }
			else
			  {
				int yy1  = y + r0 - 1;
				int nx   = ox + c0;

				for(yy = y; yy <= yy1; yy++)
				  {
					rv  =  _I[yy][ox];
					sy  -= rv;
					sy2 -= rv * rv;

					rv  =  _I[yy][nx];
					sy  += rv;
					sy2 += rv * rv;
				  }

				ox++;
		  
				di  = sy2/n - (sy * sy) / nn;
				di  = sqrt(MAX(0.0, di));
			  }
	     
			adjustment = (0.0001 + MAX(di, dt)) / (0.0001 + MIN(di,dt));

			if(adjustment < 2.0)
			  {
				// loop on template
				//
				for(j = ty0, yy = y; j <= ty1; j++, yy++)
				  for(i = tx0, xx = x; i <= tx1; i++, xx++)
					sxy += _I[yy][xx] * _T[j][i];
		  
				sxy         = (sxy - sx * sy) / (di*dt*n);
				sxy         = MAX(-1.0, MIN(sxy, 1.0));
				cvalue      = ABS(sxy);
			  }
			else
			  {
				cvalue      = 0;
			  }
	     
			if(cvalue > mr)
			  {
				mr  = cvalue;
			  }
		  }
      }

	DeleteIfTemporary(I);
	DeleteIfTemporary(T);

	return(mr);
  }


  float Correlation(array<float>& I,   // Image where the template is searched for
					array<float>& T)   // Template to be located
  {
	int		xx, x0, y0, ox = 0;
	int		yy, x1, y1, oy = 0;
	int          c0, r0;
	int		n, i, j, x, y;
	float	rv;
	float	dt, di;
	float        M[2] = {0,0};
	float        sx = 0.0, sy = 0.0, sxy = 0.0, sx2 = 0.0, sy2 = 0.0;
	int          tx0, tx1, ty0, ty1;
	float        mr = 0.0, cvalue = 0.0, nn;
	float        **_T, **_I;

	// Get some info
	//
	c0	= T.cols();
	r0	= T.rows();
   
	// and limits
	//
	x0   = I.first_col_index();
	y0   = I.first_row_index();
	x1   = I.last_col_index() - c0 + 1;
	y1   = I.last_row_index() - r0 + 1;
	tx0  = T.first_col_index();
	ty0  = T.first_row_index();
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	_T   = T.storage();
	_I   = I.storage();
   
	// Compute normalization factor for template image
	//
	sx2 = 0;
	sx  = 0;
	n   = 0;
   
	for(j = ty0; j <= ty1; j++)
      {
		for(i = tx0; i <= tx1; i++)
		  {
			sx   += _T[j][i];
			sx2  += _T[j][i] * _T[j][i];
			n++;
		  }
      }
	sx2	/= n;
	sx   /= n;
	nn    = float(n) * float(n);
   
	// Deviation of template
	//
	dt  	= sqrt(MAX((float)0.0, ((float)(sx2 - sx * sx))));
   
	// Compute correlation
	//
	for(y = y0; y <= y1; y++)
      {
		for(x = x0; x <= x1; x++)
		  {
			float adjustment;

			sxy = 0;

			if(x == x0)
			  {
				ox  = x0;

				if(y == y0)
				  {		
					int yy1 = y + r0 - 1;
					int xx1 = x + c0 - 1;
       
					sy  = sy2 = 0;
		       
					for(yy = y; yy <= yy1; yy++)
					  {
						for(xx = x; xx <= xx1; xx++)
						  {
							rv  =  _I[yy][xx];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }

					di  = (sy2/n - (sy * sy / nn));
					di  = sqrt(MAX((float)0.0, di));

					oy   = y0;
					M[0] = sy;
					M[1] = sy2;
				  }
				else
				  {
					int xx1 = x + c0 - 1;
					int ny = oy + r0;
		       
					sy  = M[0];
					sy2 = M[1];
		       
					for(xx = x; xx <= xx1; xx++)
					  {
						rv  =  _I[oy][xx];
						sy  -= rv;
						sy2 -= rv * rv;

						rv  =  _I[ny][xx];
						sy  += rv;
						sy2 += rv * rv;
					  }
		       
					oy++;
					M[0] = sy;
					M[1] = sy2;
					di   = sy2/n - (sy * sy) / nn;
					di   = sqrt(MAX((float)0.0, di));
				  }
			  }
			else
			  {
				int yy1  = y + r0 - 1;
				int nx   = ox + c0;

				for(yy = y; yy <= yy1; yy++)
				  {
					rv  =  _I[yy][ox];
					sy  -= rv;
					sy2 -= rv * rv;

					rv  =  _I[yy][nx];
					sy  += rv;
					sy2 += rv * rv;
				  }

				ox++;
		  
				di  = sy2/n - (sy * sy) / nn;
				di  = sqrt(MAX((float)0.0, di));
			  }
	     
			adjustment = (0.0001 + MAX(di, dt)) / (0.0001 + MIN(di,dt));

			if(adjustment < 2.0)
			  {
				// loop on template
				//
				for(j = ty0, yy = y; j <= ty1; j++, yy++)
				  for(i = tx0, xx = x; i <= tx1; i++, xx++)
					sxy += _I[yy][xx] * _T[j][i];
		  
				sxy         = (sxy - sx * sy) / (di*dt*n);
				sxy         = MAX((float)-1.0, MIN(sxy, (float)1.0));
				cvalue      = ABS(sxy);
			  }
			else
			  {
				cvalue      = 0;
			  }
	     
			if(cvalue > mr)
			  {
				mr  = cvalue;
			  }
		  }
      }

	DeleteIfTemporary(I);
	DeleteIfTemporary(T);
   
	return(mr);
  }


  //! This is a slightly modified version of the standard Correlation 
  //! function: the mask is used as a set of weights. Please note that
  //! the sum of the weights must be normalized to 1  by the function
  //! providing the filter
  //!
  array<double>& CorrelationW(array<double>&  II,  //! Image where the template is searched for
							  array<double>&  TT,  //! Template to be located
							  int            dd,   //! Size of neighborhood
							  int            kind, //! can be 'S' 'G' or 'L'
							  array<double>*  F,   //! Filter of image
							  array<double>*  M,   //! Mask of template
							  int           *rx,   //! Return x coordinate of best value
							  int           *ry,   //! Return y coordinate of best value
							  double         *mr   //! Return best value
							  )
  {
	array<double>	*R, T = TT, I = II;
	int		x0, y0;
	int		x1, y1;
	int          c0, c1, c2;
	int          r0, r1, r2;
	int		n, i, j, k, x, y;
	double	rv, tv, dv, rrv, ddv;
	double	dt, di, MR = -100;
	int		nn[8][2];
	double       sx, sy, sxy, sx2, sy2, su, sl;
	int          tx1, ty1;
	array<double>        *P;
	double       **_T, **_I, **_P, **_M = NULL;


	//! Make sure that the template has zero based indexes
	//!
	T.translate_at(0,0);
	if(M)
	  (*M).translate_at(0,0);
   
	//! Pixel offsets in neighborhood
	//!
	nn[0][0]	=  0;
	nn[0][1]	= -1;
	nn[1][0]	=  1;
	nn[1][1]	=  0;
	nn[2][0]	=  0;
	nn[2][1]	=  1;
	nn[3][0]	= -1;
	nn[3][1]	=  0;
	nn[4][0]	=  1;
	nn[4][1]	= -1;
	nn[5][0]	=  1;
	nn[5][1]	=  1;
	nn[6][0]	= -1;
	nn[6][1]	=  1;
	nn[7][0]	= -1;
	nn[7][1]	= -1;
   
   
	//! Get some info
	//!
	c0	= T.cols();
	r0	= T.rows();
	c2	= I.cols();
	r2	= I.rows();
   
	//! Correlation region size
	//!
	c1	= c2 - c0 + 1;
	r1	= r2 - r0 + 1;
   
	//! and limits
	//!
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<double>(y0, y1, x0, x1);
	P    = new array<double>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();

	if(M)
	  _M   = (*M).storage();

   
	//! Compute normalization factor for template image
	//!
	sx2 = 0;
	sx  = 0;
	n   = 0;

	if(M == NULL)
	  {
		for(j = 0; j <= ty1; j++)
		  {
			for(i = 0; i <= tx1; i++)
			  {
				sx   += _T[j][i];
				sx2  += _T[j][i] * _T[j][i];
				n++;
			  }
		  }

		sx2	/= n;
		sx   /= n;
	  }
	else
	  {
		for(j = 0; j <= ty1; j++)
		  {
			for(i = 0; i <= tx1; i++)
			  {
				if(_M[j][i] > 0)
				  {
					sx   += _M[j][i] * _T[j][i];
					sx2  += _M[j][i] * _T[j][i] * _T[j][i];
					n++;
				  }
			  }
		  }
	  }
   
	//! Deviation of template
	//!
	dt  	= sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
	//! Normalize template to zero average and unit sigma
	for(j = 0; j <= ty1; j++)
	  for(i = 0; i <= tx1; i++)
		_T[j][i] = (_T[j][i] - sx) / dt;
   
   
	//! Compute correlation
	//!
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
	  {
		for(x = x0; x <= x1; x++)
		  {
			int   nk = 2*dd+tx1+1;
			double w;
	     
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				//! Copy image patch into patch area
				//!
				for(j = -dd; j <= dd+ty1; j++)
				  memcpy(&(_P[j][-dd]), &(_I[y+j][x-dd]), sizeof(double)*nk);

				//! Reset accumulators
				//!
				sxy = 0;
				sy  = 0;
				sy2 = 0;
				su  = 0;
				sl  = 0;
		  
				//! Normalize image patch
				//!
				if(M == NULL)
				  {
					for(j =  0; j <= ty1; j++)
					  {
						for(i = 0; i <= tx1; i++)
						  {
							rv  =  _I[y+j][x+i];
							sy  += rv;
							sy2 += rv * rv;
						  }
					  }
					sy  /= n;
					sy2 /= n;
				  }
				else
				  {
					for(j =  0; j <= ty1; j++)
					  {
						for(i = 0; i <= tx1; i++)
						  {
							rv  =  _I[y+j][x+i];
							sy  += _M[j][i] * rv;
							sy2 += _M[j][i] * rv * rv;
						  }
					  }
				  }
		  
				di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
		  
				for(j = -dd; j <= dd+ty1; j++)
				  for(i = -dd; i <= dd+tx1; i++)
					_P[j][i] = (_P[j][i] - sy) / di;
		  
		  
				//! loop on template
				//!
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if(M == NULL)
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  }
						  } else if ((w = _M[j][i]) > 0) {
							rv =  _P[j][i];
							tv =  _T[j][i];
				 
							//! look for most similar pixel in 
							//! the neighborhood of radius 1 
							//! using 4-connectedness
							//!
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }
				 
							switch(kind)
							  {
							  case 'S':
								sxy += w * tv * rv;
								break;
							  case 'G':
								su  += w * ABS(tv - rv);
								sl  += w * (ABS(tv) + ABS(rv));
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += w * ABS(tv - rv) / sl;
								break;
							  }			       
						  }
			    
					  }
				  }

				switch(kind)
				  {
				  case 'S':
					sxy         = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					(*R)[y][x]  = 100*(1.0 - su);
					break;
				  }
		  
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
	  }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);
   
	if(M != NULL)
	  DeleteIfTemporary(*M);
   
	//! Adjust size 
	//!
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	// We want to perform it at R level
	// *mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }


  // This is a slightly modified version of the standard Correlation 
  // function: the mask is used as a set of weights. Please note that
  // the sum of the weights must be normalized to 1  by the function
  // providing the filter
  //
  array<double>& CorrelationW2(array<double>&  II,   // Image where the template
                                                     // is searched for
                               array<double>&  TT,   // Template to be located
                               int             dd,   // Size of neighborhood
                               int             kind, // can be 'S' 'G' 'L' 'P'
                               int             normalize, // 0/1
                               int             p,    // used to determine L_p
                               double          maxD, // used to determine L_p
                               array<double>*  F,    // Filter of image
                               array<double>*  M,    // Mask of template
                               int            *rx,   // Best value x
                               int            *ry,   // Best value y
                               double         *mr    // Best value
							   )
  {
	array<double>    *R, T = TT, I = II;
	int             x0, y0;
	int             x1, y1;
	int             c0, c1, c2;
	int             r0, r1, r2;
	int             n, i, j, k, x, y;
	double          rv, tv, dv, rrv, ddv;
	double          dt, di, MR = -100;
	int             nn[8][2];
	double          sx, sy, sxy, sx2, sy2, su, sl, sp;
	int             tx1, ty1;
	array<double>  *P;
	double        **_T, **_I, **_P, **_M = NULL;

	// Make sure that the template has zero based indexes
	//
	T.translate_at(0,0);
	if(M)
	  (*M).translate_at(0,0);
   
	// Pixel offsets in neighborhood
	//
	nn[0][0] =  0;
	nn[0][1] = -1;
	nn[1][0] =  1;
	nn[1][1] =  0;
	nn[2][0] =  0;
	nn[2][1] =  1;
	nn[3][0] = -1;
	nn[3][1] =  0;
	nn[4][0] =  1;
	nn[4][1] = -1;
	nn[5][0] =  1;
	nn[5][1] =  1;
	nn[6][0] = -1;
	nn[6][1] =  1;
	nn[7][0] = -1;
	nn[7][1] = -1;
   
   
	// Get some info
	//
	c0   = T.cols();
	r0   = T.rows();
	c2   = I.cols();
	r2   = I.rows();
   
	// Correlation region size
	//
	c1   = c2 - c0 + 1;
	r1   = r2 - r0 + 1;
   
	// and limits
	//
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<double>(y0, y1, x0, x1);
	P    = new array<double>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();

	if(M)
	  _M   = (*M).storage();

   
	// Compute normalization factor for template image
	//
	sx2 = 0;
	sx  = 0;
	n   = 0;

	if(normalize == 1) 
	  {
		if(M == NULL)
		  {
			n = T.rows() * T.cols();

			for(j = 0; j <= ty1; j++)
			  {
				for(i = 0; i <= tx1; i++)
				  {
					sx   += _T[j][i];
					sx2  += _T[j][i] * _T[j][i];
				  }
			  }

			sx2  /= n;
			sx   /= n;
		  }
		else
		  {
			for(j = 0; j <= ty1; j++)
			  {
				for(i = 0; i <= tx1; i++)
				  {
					if(_M[j][i] > 0)
					  {
						sx   += _M[j][i] * _T[j][i];
						sx2  += _M[j][i] * _T[j][i] * _T[j][i];
						n++;
					  }
				  }
			  }
		  }
   
		// Deviation of template
		//
		dt   = sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
		// Normalize template to zero average and unit sigma
		if(dt > 0.0)
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				_T[j][i] = (_T[j][i] - sx) / dt;
		  }
		else 
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				_T[j][i] = 0;
		  }
	  } 
	else 
	  {
		if(M == NULL)
		  {
			n = T.rows() * T.cols();
		  }
		else
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				if(_M[j][i] > 0)
				  n++;
		  }
	  }
   
	// Compute correlation
	//
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
	  {
		for(x = x0; x <= x1; x++)
		  {
			int   nk = 2*dd+tx1+1;
			double w;
         
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				// Copy image patch into patch area
				//
				for(j = -dd; j <= dd+ty1; j++)
				  memcpy(&(_P[j][-dd]), &(_I[y+j][x-dd]), sizeof(double)*nk);

				// Reset accumulators
				//
				sxy = 0;
				su  = 0;
				sl  = 0;
				sp  = 0;

				if(normalize == 1)
				  {
					sy  = 0;
					sy2 = 0;
					// Normalize image patch
					//
					if(M == NULL)
					  {
						for(j =  0; j <= ty1; j++)
						  {
							for(i = 0; i <= tx1; i++)
							  {
								rv  =  _I[y+j][x+i];
								sy  += rv;
								sy2 += rv * rv;
							  }
						  }
						sy  /= n;
						sy2 /= n;
					  }
					else
					  {
						for(j =  0; j <= ty1; j++)
						  {
							for(i = 0; i <= tx1; i++)
							  {
								rv  =  _I[y+j][x+i];
								sy  += _M[j][i] * rv;
								sy2 += _M[j][i] * rv * rv;
							  }
						  }
					  }
          
					di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
          
					if(di > 0.0)
					  {
						for(j = -dd; j <= dd+ty1; j++)
						  for(i = -dd; i <= dd+tx1; i++)
							_P[j][i] = (_P[j][i] - sy) / di;
					  }
					else 
					  {
						for(j = -dd; j <= dd+ty1; j++)
						  for(i = -dd; i <= dd+tx1; i++)
							_P[j][i] = 0;
					  }
				  }
          
				// loop on template
				//
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if(M == NULL)
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
                 
							// look for most similar pixel in 
							// the neighborhood of radius 1 
							// using 4-connectedness
							//
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)

								  su += ABS(tv - rv) / sl;
								break;
							  case 'P':
								sp += pow(ABS(tv - rv), p);
								break;
							  }
						  } else if ((w = _M[j][i]) > 0) {
							rv =  _P[j][i];
							tv =  _T[j][i];
                 
							// look for most similar pixel in 
							// the neighborhood of radius 1 
							// using 4-connectedness
							//
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }
                 
							switch(kind)
							  {
							  case 'S':
								sxy += w * tv * rv;
								break;
							  case 'G':
								su  += w * ABS(tv - rv);
								sl  += w * (ABS(tv) + ABS(rv));
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += w * ABS(tv - rv) / sl;
								break;
							  case 'P':
								sp += w * pow(ABS(tv - rv), p);
								break;
							  }                 
						  }
                
					  }
				  }

				switch(kind)
				  {
				  case 'S':
					if(M == NULL)
					  sxy /= n;
					sxy         = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					if(M == NULL)
					  su /= n;
					(*R)[y][x]  = 100*(1.0 - su);
					break;
				  case 'P':
					if(M == NULL) 
					  sp /= n;
				   
					sp = 1.0 - pow(sp, 1.0/p) / maxD;
					(*R)[y][x]  = 100*sp;
					break;
				  }
          
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
	  }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);
   
	if(M != NULL)
	  DeleteIfTemporary(*M);
   
	// Adjust size 
	//
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	// We want to perform it at R level
	// *mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }


array<float>& CorrelationW2(array<float>&  II,        // Image where the template is searched for
							array<float>&  TT,        // Template to be located
							int            dd,        // Size of neighborhood
							int            kind,      // can be 'S' 'G' 'L' 'P'
							int            normalize, // 0/1
							int            p,         // used to determine L_p
							float          maxD,      // used to determine L_p
							array<float>*  F,         // Filter of image
							array<float>*  M,         // Mask of template
							int            *rx,       // Best value x
							int            *ry,       // Best value y
							float          *mr        // Best value
							)
  {
	array<float>    *R, T = TT, I = II;
	int             x0, y0;
	int             x1, y1;
	int             c0, c1, c2;
	int             r0, r1, r2;
	int             n, i, j, k, x, y;
	double          rv, tv, dv, rrv, ddv;
	double          dt, di, MR = -100;
	int             nn[8][2];
	double          sx, sy, sxy, sx2, sy2, su, sl, sp;
	int             tx1, ty1;
	array<float>  *P;
	float        **_T, **_I, **_P, **_M = NULL;

	// Make sure that the template has zero based indexes
	//
	T.translate_at(0,0);
	if(M)
	  (*M).translate_at(0,0);
   
	// Pixel offsets in neighborhood
	//
	nn[0][0] =  0;
	nn[0][1] = -1;
	nn[1][0] =  1;
	nn[1][1] =  0;
	nn[2][0] =  0;
	nn[2][1] =  1;
	nn[3][0] = -1;
	nn[3][1] =  0;
	nn[4][0] =  1;
	nn[4][1] = -1;
	nn[5][0] =  1;
	nn[5][1] =  1;
	nn[6][0] = -1;
	nn[6][1] =  1;
	nn[7][0] = -1;
	nn[7][1] = -1;
   
   
	// Get some info
	//
	c0   = T.cols();
	r0   = T.rows();
	c2   = I.cols();
	r2   = I.rows();
   
	// Correlation region size
	//
	c1   = c2 - c0 + 1;
	r1   = r2 - r0 + 1;
   
	// and limits
	//
	x0   = I.first_col_index() + dd;
	y0   = I.first_row_index() + dd;
	x1   = I.last_col_index()  - dd - c0 + 1;
	y1   = I.last_row_index()  - dd - r0 + 1;
	tx1  = T.last_col_index();
	ty1  = T.last_row_index();
   
	R    = new array<float>(y0, y1, x0, x1);
	P    = new array<float>(-dd,+dd+ty1, -dd, +dd+tx1);


	_T   = T.storage();
	_I   = I.storage();
	_P   = (*P).storage();

	if(M)
	  _M   = (*M).storage();

   
	// Compute normalization factor for template image
	//
	sx2 = 0;
	sx  = 0;
	n   = 0;

	if(normalize == 1) 
	  {
		if(M == NULL)
		  {
			n = T.rows() * T.cols();

			for(j = 0; j <= ty1; j++)
			  {
				for(i = 0; i <= tx1; i++)
				  {
					sx   += _T[j][i];
					sx2  += _T[j][i] * _T[j][i];
				  }
			  }

			sx2  /= n;
			sx   /= n;
		  }
		else
		  {
			for(j = 0; j <= ty1; j++)
			  {
				for(i = 0; i <= tx1; i++)
				  {
					if(_M[j][i] > 0)
					  {
						sx   += _M[j][i] * _T[j][i];
						sx2  += _M[j][i] * _T[j][i] * _T[j][i];
						n++;
					  }
				  }
			  }
		  }
   
		// Deviation of template
		//
		dt   = sqrt(MAX(0.0, ((double)(sx2 - sx * sx))));
   
		// Normalize template to zero average and unit sigma
		if(dt > 0.0)
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				_T[j][i] = (_T[j][i] - sx) / dt;
		  }
		else 
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				_T[j][i] = 0;
		  }
	  } 
	else 
	  {
		if(M == NULL)
		  {
			n = T.rows() * T.cols();
		  }
		else
		  {
			for(j = 0; j <= ty1; j++)
			  for(i = 0; i <= tx1; i++)
				if(_M[j][i] > 0)
				  n++;
		  }
	  }
   
	// Compute correlation
	//
	*rx =  x0;
	*ry =  y0;
	for(y = y0; y <= y1; y++)
	  {
		for(x = x0; x <= x1; x++)
		  {
			int   nk = 2*dd+tx1+1;
			float w;
         
			if((F == NULL) || ((*F)[y][x] > 0))
			  {
				// Copy image patch into patch area
				//
				for(j = -dd; j <= dd+ty1; j++)
				  memcpy(&(_P[j][-dd]), &(_I[y+j][x-dd]), sizeof(float)*nk);

				// Reset accumulators
				//
				sxy = 0;
				su  = 0;
				sl  = 0;
				sp  = 0;

				if(normalize == 1)
				  {
					sy  = 0;
					sy2 = 0;
					// Normalize image patch
					//
					if(M == NULL)
					  {
						for(j =  0; j <= ty1; j++)
						  {
							for(i = 0; i <= tx1; i++)
							  {
								rv  =  _I[y+j][x+i];
								sy  += rv;
								sy2 += rv * rv;
							  }
						  }
						sy  /= n;
						sy2 /= n;
					  }
					else
					  {
						for(j =  0; j <= ty1; j++)
						  {
							for(i = 0; i <= tx1; i++)
							  {
								rv  =  _I[y+j][x+i];
								sy  += _M[j][i] * rv;
								sy2 += _M[j][i] * rv * rv;
							  }
						  }
					  }
          
					di  = sqrt(MAX(0.0, ((double)(sy2 - sy * sy))));
          
					if(di > 0.0)
					  {
						for(j = -dd; j <= dd+ty1; j++)
						  for(i = -dd; i <= dd+tx1; i++)
							_P[j][i] = (_P[j][i] - sy) / di;
					  }
					else 
					  {
						for(j = -dd; j <= dd+ty1; j++)
						  for(i = -dd; i <= dd+tx1; i++)
							_P[j][i] = 0;
					  }
				  }
          
				// loop on template
				//
				for(j =  0; j <= ty1; j++)
				  {
					for(i = 0; i <= tx1; i++)
					  {
						if(M == NULL)
						  {
							rv =  _P[j][i];
							tv =  _T[j][i];
                 
							// look for most similar pixel in 
							// the neighborhood of radius 1 
							// using 4-connectedness
							//
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }

							switch(kind)
							  {
							  case 'S':
								sxy += tv * rv;
								break;
							  case 'G':
								su  += ABS(tv - rv);
								sl  += ABS(tv) + ABS(rv);
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += ABS(tv - rv) / sl;
								break;
							  case 'P':
								sp += pow(ABS(tv - rv), p);
								break;
							  }
						  } else if ((w = _M[j][i]) > 0) {
							rv =  _P[j][i];
							tv =  _T[j][i];
                 
							// look for most similar pixel in 
							// the neighborhood of radius 1 
							// using 4-connectedness
							//
							if(dd != 0)
							  {
								dv =  ABS(tv - rv);
								for(k = 0; k < 4; k++)
								  {
									rrv = _P[j+nn[k][0]][i+nn[k][1]];
									ddv = ABS(tv - rrv);
									if(ddv < dv)
									  {
										dv = ddv;
										rv = rrv;
									  }
								  }
							  }
                 
							switch(kind)
							  {
							  case 'S':
								sxy += w * tv * rv;
								break;
							  case 'G':
								su  += w * ABS(tv - rv);
								sl  += w * (ABS(tv) + ABS(rv));
								break;
							  case 'L':
								sl  = ABS(tv) + ABS(rv);
								if(sl > 0)
								  su += w * ABS(tv - rv) / sl;
								break;
							  case 'P':
								sp += w * pow(ABS(tv - rv), p);
								break;
							  }                 
						  }
                
					  }
				  }

				switch(kind)
				  {
				  case 'S':
					if(M == NULL)
					  sxy /= n;
					sxy         = MAX(-1.0, MIN(sxy, 1.0));
					(*R)[y][x]  = 100*sxy;
					break;
				  case 'G':
					(*R)[y][x]  = 100*(1.0 - su / sl);
					break;
				  case 'L':
					if(M == NULL)
					  su /= n;
					(*R)[y][x]  = 100*(1.0 - su);
					break;
				  case 'P':
					if(M == NULL) 
					  sp /= n;
				   
					sp = 1.0 - pow(sp, 1.0/p) / maxD;
					(*R)[y][x]  = 100*sp;
					break;
				  }
          
				if((*R)[y][x] > MR)
				  {
					MR  = (*R)[y][x];
					*mr = (MR/100.0);
					*rx = x;
					*ry = y;
				  }
			  }
		  }
	  }
   
	delete P;
   
	if(F != NULL)
	  DeleteIfTemporary(*F);
   
	if(M != NULL)
	  DeleteIfTemporary(*M);
   
	// Adjust size 
	//
	(*R) = (*R).subimage(y0-dd, y1+dd, x0-dd, x1+dd);
   
	// We want to perform it at R level
	// *mr = RemapCorrelationValue(*mr, kind);
   
	return((*R).temporary());
  }

  array<float>& HierarchicalCorrelationW2(array<float>*   image_pyramid,
										  array<float>*   template_pyramid,
										  array<float>*   filter_pyramid,
										  array<float>*   mask_pyramid,
										  int             levels,
										  int*            neighborhood,
										  int             kind, 
										  int             p,
										  double          maxD,
										  double          fraction_to_keep,
										  int            *rx, 
										  int            *ry, 
										  double         *mr,  
										  vector<double> *bvs,
										  int             unusedLevels
										)
  {
	array<float>    *correlation_map;
	array<float>    *filter;
	array<float>    *R;
	int      best_value_x, best_value_y;
	int      n;
	int      i, l;
	int      x, y;
	int      x0, x1;
	int      y0, y1;
	float   best_value;
	float   dont_keep;
	float   minimum_value_to_keep;
	vector<float>   histogram(101);
	float   **_f;
   
	correlation_map = new array<float>[levels];
	filter          = new array<float>[levels];
   
	if(filter_pyramid == NULL) 
	  {
		filter[0] = (* (new array<float>(image_pyramid[0].first_row_index(),
										 image_pyramid[0].last_row_index(),
										 image_pyramid[0].first_col_index(),
										 image_pyramid[0].last_col_index()))).temporary();
		filter[0] = 1.0;
	  }
	else 
	  {
		filter[0] = filter_pyramid[0] > 0.0;
	  }

	_f        = filter[0].storage();
   
	for(l = 0; l < levels - unusedLevels; l++)
      {
		float **_C;
	
		//! Compute correlation map at the given level 
		//! focussing computation on the sites which
		//! were found to be promising at the 
		//! previous level
		//!
		{
		  array<float> R, T, B;
	   
		  R = image_pyramid[l];
		  T = template_pyramid[l];
		  B = mask_pyramid[l];

		  correlation_map[l] = CorrelationW2(R,
											 T,
											 neighborhood[l],
											 kind, 1, p, maxD,
											 &(filter[l]),
											 &B,
											 &best_value_x, &best_value_y, &best_value);

					
		  if(bvs)
			(*bvs)[l] = best_value;
		}
	
		_C = correlation_map[l].storage();
	
		//! Reset histogram
		//!
		histogram =  histogram * 0;
	
		//! Scan correlation map to compute the distribution
		//! of the correlation values
		//!
		x0 = correlation_map[l].first_col_index();
		x1 = correlation_map[l].last_col_index();
		y0 = correlation_map[l].first_row_index();
		y1 = correlation_map[l].last_row_index();
	
	
		//! Normalize correlation values to the interval [0,100]
		//!
		{
		  float  max_c_value = FLT_MIN, min_c_value  = FLT_MAX;
		  int    init = TRUE;
	   
		  n = 0;
		  for(y = y0; y <= y1; y++)
			for(x = x0; x <= x1; x++)
			  if(filter[l][y][x] > 0.0)
				{
				  if(init == TRUE)
					{
					  init = FALSE;
					  max_c_value = min_c_value = correlation_map[l][y][x];
					}
				  else
					{
					  max_c_value = MAX(max_c_value, correlation_map[l][y][x]);
					  min_c_value = MIN(min_c_value, correlation_map[l][y][x]);
					}
				  n++;
				}
	   
		  if(n > 0)
			{
			  for(y = y0; y <= y1; y++)
				for(x = x0; x <= x1; x++)
				  if(filter[l][y][x] > 0.0)
					{
					  correlation_map[l][y][x] = 
						((correlation_map[l][y][x]-min_c_value) /
						 (max_c_value - min_c_value))*100.0;
					}
			}
		}
	
		for(y = y0; y <= y1; y++)
		  for(x = x0; x <= x1; x++)
			if(filter[l][y][x] > 0.0)
			  {
				histogram[int(correlation_map[l][y][x])]++;
				n++;
			  }
	
		//! Determine the minimum correlation value for
		//! which we want to compute correlation at
		//! the next level
		//!
		dont_keep = n*(1-fraction_to_keep);
		n         = 0;
		i         = 0;
		while((n <= dont_keep) && (i <= 100))
		  n += int(histogram[i++]);
	
		minimum_value_to_keep = MAX(0,(i-1));
	
		//! Compute the filter to be used at the next level
		//!
		if(l+1 < levels - unusedLevels)
		  {
			x0 = image_pyramid[l+1].first_col_index();
			x1 = image_pyramid[l+1].last_col_index();
			y0 = image_pyramid[l+1].first_row_index();
			y1 = image_pyramid[l+1].last_row_index();
	     
			filter[l+1] = correlation_map[l] > minimum_value_to_keep - 1;

			filter[l+1] = Grow(filter[l+1], 1);
			filter[l+1] = Zoom(filter[l+1], 2, 2);

			//! Make sure that dimensions match
			//!
			filter[l+1] = filter[l+1].subimage(y0, y1, x0, x1);

			if(filter_pyramid != NULL)
			  filter[l+1] = filter[l+1] & (filter_pyramid[l+1] > 0).subimage(y0, y1, x0, x1);
		  }
		else
		  {
			*rx = best_value_x;
			*ry = best_value_y;

			//! and scale appropriately
			//!
			for(int i = 0; i < unusedLevels; i++)
			  {
				*rx *= 2;
				*ry *= 2;
			  }
	     
			*mr = best_value;
		  }
      }
   
	R = new array<float>(correlation_map[levels-1]);
   
	delete [] correlation_map;
	delete [] filter;
   
	*mr = RemapCorrelationValue(*mr, kind);

	return((*R).temporary());
  }


}
