#ifndef LINEAR_H
#define LINEAR_H

#include <arrayTools/Array.h>
#include <arrayTools/Linear_base.h>

namespace _ARRAYTOOLS_NAMESPACE
{

  //====================================================================================================
  // GAUSS JORDAN
  //====================================================================================================

  array<double> & Gaussj(array<double> & A,          /* matrix n x n                */
			 array<double> & B,       /* matrix of rhs vectors n x m */
			 array<double> & InvA     /* overwrited: inverse matrix  */ );

  array<double> & Gaussj(array<double> & A,          /* matrix n x n                */
			 array<double> & B        /* matrix of rhs vectors n x m */ );

 //====================================================================================================
  // HOUSEHOLDER AND TQLI
  //====================================================================================================

  //! Householder reduction of a real symmetric matrix in tridiagonal form
  void Householder(array<double>  &  A, // input unchanged del if temp not overwrited
		   array<double>  &  Q, // output
		   vector<double> &  D, // output
		   vector<double> &  E);// output


  //! QL algorithm with implicit shifts to  determine eval and evect of a real symmetric tridiagonal matrix
  void Tqli(vector<double> &  D,   //input del if temp not overwrited
	    vector<double> &  E,   //input del if temp not overwrited
	    array<double>  &  Q,   //input del if temp not overwrited
	    vector<double> &  EV,  //output eigenvalues
	    array<double>  &  Z);  //output eigenvectors 




  //! Perform tridiagonal decomposition and ql algorithm to determine eval and evect
  //! D = Tr(Z) * A * Z
  //! So the cols of Z are the eigenvectors!!
  void HouseholderAndTqli(array<double>  &  A,   // input unchanged del if temp not overwrited
			  vector<double> &  EV,  //output eigenvalues
			  array<double>  &  Z);  //output eigenvectors 

  //====================================================================================================
  // LU 
  //====================================================================================================


  /*!
   * Linear Equation solver via LU decomposition
   * Ax=B
   * where A is a nxn matrix
   * B is a matrix of vectors(columns) of nxm so we have m solutions
   * It returns an array of dim nxm of solutions
   */
  array<double> & LU_LinearEq(array<double> & A,  /* matrix n x n                */
			      array<double> & B   /* matrix of rhs vectors n x m */ );




  //====================================================================================================
  // SVD 
  //====================================================================================================

  /*!
   * SV Decomposition with arrays and vectors 
   * A = U * W * tr(V) 
   * with A and U m x n, W diagonal (vector) nxn and V nxn
   * the input matrix A m x n is overwrited with U
   */

  void svdcmp(array<double>  & A,  /* input matrix m x n overwrited with U*/
	      vector<double> & W,  /* diagonal of a matrix  n x n */
	      array<double>  & V   /* n x n matrix */ );




  /*!
   * With array and vectors
   * Backsubstitution: it solves A*X = B
   * A (m x n)
   * B is a vector of dim m
   * X is the solution, a vector of dim n
   * A is already decomposed in U W V
   */


  vector<double> & svbksb(array<double> &  U,
			      vector<double>&  W,
			      array<double> &  V,
			      vector<double> & B);

  /*!
   * With array and vectors
   * Linear Equation solver via SV decomposition
   * A * X = B
   * A matrix m x n 
   * B vector rhs vector of dim m
   * X solution overwrited, vector of dim n 
   * it perform a thresholding if the singular value w is
   * w < wmax * threshold => w = 0
   */
  vector<double> & SV_LinearEq(array<double>  &   A,  /* matrix m x n                         */
				   vector<double> &   B,  /* vector rhs vector of dim m           */
				   double threshold =  1.0e-12);


 

  template<class T> array<T>&	LU_Inverse(array<T>& AA)
  {
    int		        *indx;
    int		        n = AA.rows();
    int		        i, j;

    double*	        col;
    double*               b;
    double	        d;

    array<double>         A(CastTo<double>(AA));
    array<double>         B; //LUD decomposition	
     
    double**              _A;
    double**              _B;
    
    if (n != AA.cols())
      {
	fprintf(stderr,"Inverse:\nCannot invert rectangular matrix!\n");
	exit(-1);
      }
  
    // Allocate vector to keep track of row permutation
    indx	= new int[n];

    // Allocate aux vectors
    col = new double[n];
    b   = new double[n];
      
    B  = A;
    _A = A.storage();
    _B = B.storage();
    

    // Compute LU decomposition 
    if( ludcmp(_B, indx, n, &d) == FALSE )
      {
	return (new array<T>)->temporary();
      }

    array<T> *I = new array<T>(n,n);
    T ** _I = I->storage();

    for(j = 0; j < n; j++)
      {
      
	// Fill the j-th column 
	for(i = 0; i < n; i++)
	  {
	    b[i]   = 0.0;
	    col[i] = 0.0;
	  }
	b[j]   = 1.0;
	col[j] = 1.0;
      
	lubksb(_B, indx, col, n);
  
	//Solution refinement    
	mprove(_A, _B, indx, b, col, n);
	mprove(_A, _B, indx, b, col, n);
     
	//Copy the solution
	for(i = 0; i < n; i++)
	  _I[i][j] = (T) col[i];
      }
    
    delete[] indx;
    delete[] col;
    delete[] b;

    return I->temporary();
  }


  template<class T> array<T> & SV_Inverse(array<T> & A, double threshold =  1.0e-12 )
  {
  
    int		        n = A.rows();
    int		        i, j;
 

    if (n != A.cols())
      {
	fprintf(stderr,"SV_Inverse:\nCannot invert rectangular matrix!\n");
	exit(-1);
      }
  
    //SV decomposition
    array<double> U(CastTo<double>(A)); 	
    array<double> V(n,n);
    array<T> * I = new array<T>(n,n);  

    double**              _U = U.storage();
    double*               _W = new double[n];
    double**              _V = V.storage();
    T     **              _I = (*I).storage();
  

    double * b = new double[n];
    double * x = new double[n];

 

  // Compute SV decomposition 
  svdcmp(_U, _W, _V, n, n);
  
  // Threshold
  svd_w_threshold(_W, n, threshold);

  for(j = 0; j < n; j++)
    {
      
      // Fill the j-th column 
      for(i = 0; i < n; i++)
	{
	b[i]   = 0.0;
	}
      b[j]   = 1.0;
      
      svbksb(_U, _W, _V, b, x, n, n);
  
      
     
      //Copy the solution
      for(i = 0; i < n; i++)
	_I[i][j] = (T) x[i];
    }
    
  delete[] _W;
  delete[] b;
  delete[] x;

  return (*I).temporary();
}



}//namespace




#endif
