#ifndef LINEAR_BASE_H
#define LINEAR_BASE_H


namespace _ARRAYTOOLS_NAMESPACE
{

  /*!
   * SV Decomposition 
   * A = U * W * tr(V) 
   * with A and U m x n, W diagonal (vector) nxn and V nxn
   * the input matrix A m x n is overwrited with U
   */

  void svdcmp(double**  a, /* input matrix m x n overwrited with U*/
	      double*   w, /* diagonal of a matrix  n x n */
	      double**  v, /* n x n matrix */
	      int       m,
	      int       n);

  /*!
   * Backsubstitution: it solves A*X = B
   * A (m x n)
   * B is a vector of dim m
   * X is the solution, a vector of dim n
   * A is already decomposed in U W V
   */

  void svbksb(double**  u,
	      double*   w,
	      double**  v, /* From svdcmp */
	      double*   b, /* vector dim = m */
	      double*   x, /* output dim = n */ 
	      int       m,
	      int       n);



  // Threshold s values
  // w[n] = 0 if w[n] < max*threshold
  void svd_w_threshold(double * W, int n, double threshold);
 
  /*!
   * Linear Equation solver via SV decomposition
   * A * X = B
   * A matrix m x n 
   * B vector rhs vector of dim m
   * X solution overwrited, vector of dim n 
   * it perform a thresholding if the singular value w is
   * w < wmax * threshold => w = 0
   */
  void svlneq(double **  A,  /* matrix m x n                         */
	      double *   B,  /* vector rhs vector of dim m           */
	      double *   X,  /* solution overwrited, vector of dim n */
	      int m,
	      int n,
	      double threshold =  1.0e-12);

  /*!
   * Replace a with the LUD decomposition
   * indx keep record of permutations
   * d is 1 if permutations are even -1 if odd
   */
  int                 ludcmp(double** 	 a,
			     int*              indx,
			     int               n,
			     double*           d);

  /*!
   * Backsubstitution
   * Solve AX=B
   * a is the LUD decomposition
   * b is the rhs vector and is replaced by x  
   */
  void                 lubksb(double**  	 a,
			      int*              indx,
			      double*	         b,
			      int               n);

  /*!
   * Iterative improvement of solution
   * x is refined 
   */
  void                 mprove(double**  a, 
			      double**  alud, 
			      int*      indx,  
			      double*   b, 
			      double*   x,
			      int       n);
  /*!
   * Linear Equation solver via LU decomposition
   * A matrix n x n 
   * B,vector rhs vector of dim n
   * X,solution overwrited, vector of dim n 
   *
   * return FALSE if no solutions
   */
  int lulneq(double **  A,  /* matrix n x n                         */
	     double *   B,  /* vector rhs vector of dim n           */
	     double *   X,  /* solution overwrited, vector of dim n */
	     int n);




  //! Householder reduction of a real symmetric matrix in tridiagonal form
  void householder(double ** a, // nxn input matrix, overwrited with the matrix that perform the transformation
		   double *  d, // vector of diagonal elements
		   double *  e, // vector of off diagonal elements
		   int n);      // dimension


  //! TQLI algrithm:
  //! QL algorithm with implicit shifts to  determine eval and evect of a real symmetric tridiagonal matrix
  void tqli(double *  d, // diagonal elements    -> eigenvalues
	    double *  e, // offdiagonal elements -> destroyed
	    double ** z, 
	    // tranform matrix q    -> normalized eigenvectors of a
	    // identity             -> eigenvectors of the tridiagonal matrix
	    int       n);






}


#endif
