#ifndef UTILS_H
#define UTILS_H

#include <cmath>
#include <arrayTools/Array.h>

namespace _ARRAYTOOLS_NAMESPACE
{


  vector<double> & Histogram(array<unsigned char> & A, int bins);

  template<class T> void MinMax(array<T> & A, T & minimum, T & maximum)
    {
  
      double max, min;
      int    r0,r1,c0,c1;
      int    i,j;
      T      v;  

      max = -HUGE_VAL;
      min = HUGE_VAL;
  
      c0           = A.first_col_index();
      c1           = A.last_col_index();
      r0           = A.first_row_index();
      r1           = A.last_row_index();

      T ** _A          =  A.storage();

      for(i = r0; i <= r1; i++)
	{
	  for(j = c0; j <= c1; j++)
	    {
	      v =  _A[i][j] ; 
	      min = (((min) < (v))? (min): (v));
	      max = (((max) > (v))? (max): (v));
	    }
	}

  
      minimum = (T) min;
      maximum = (T) max;
  

    }

  template<class T> void MinMax(vector<T> & A, T & minimum, T & maximum)
    {
  
      double max, min;
      int i0, i1;
      int    i;
      T      v;  

      max = -HUGE_VAL;
      min = HUGE_VAL;
  
      i0           = A.first_index();
      i1           = A.last_index();

      T * _A          =  A.storage();

      for(i = i0; i <= i1; i++)
	{
	  v =  _A[i] ; 
	  min = (((min) < (v))? (min): (v));
	  max = (((max) > (v))? (max): (v));
	}

  
      minimum = (T) min;
      maximum = (T) max;
    }


  //! Energy of image - subimage
  double Image_Energy(double ** _A, int r0, int r1, int c0, int c1);

  //! Energy of image
  double Image_Energy(array<double> & A);



  //! Rescale form [bottom, top) 
  //! max = top - 1.0e-10
  array<double> & LinearRescaleDouble(array<double> & A, double top = 1.0, double bottom = 0.0);



  //! Histogram equalization
  array<unsigned char> & HistEq(array<unsigned char> & A);

  //! Histogram eq double: precision is the n of quanta
  array<double> & HistEq(array<double> & A, int precision);

  //! Log transform of image of double
  array<double> & LogaritmicTransform(array<double> & A);



  //! Average level of image
  template<class T> double Average(array<T> & A)
    {
  
      double avg;
      int    r0,r1,c0,c1;
      int    i,j;
  
      c0           = A.first_col_index();
      c1           = A.last_col_index();
      r0           = A.first_row_index();
      r1           = A.last_row_index();

      T ** _A          =  A.storage();

      avg = 0.0;
      for(i = r0; i <= r1; i++)
	{
	  for(j = c0; j <= c1; j++)
	    {
	      avg +=  (double)_A[i][j] ; 
	    }
	}
  
      avg /= (A.rows() * A.cols());
  
      return avg;
    }


  //! Template for unsigned char rescale [0, 255]
  template<class T> array<unsigned char> & LinearRescale(array<T> & A)
    {
  
      double max,min,scale,offset;
      int    r0,r1,c0,c1;
      int    i,j;
      double   v;  

      max = -MAXDOUBLE;
      min = MAXDOUBLE;
  
      c0           = A.first_col_index();
      c1           = A.last_col_index();
      r0           = A.first_row_index();
      r1           = A.last_row_index();

      array<unsigned char> * R =  new array<unsigned char>(r0,r1,c0,c1);
      unsigned char ** _R      =  (*R).storage();
      T ** _A                  =  A.storage();

      for(i = r0; i <= r1; i++)
	{
	  for(j = c0; j <= c1; j++)
	    {
	      v   = (double)( _A[i][j] );  
	      min = ( (min < v) ? min : v);
	      max = ( (max > v) ? max : v);
	    }
	}
  
      scale    = (256 - 1.0e-10) / (max - min);
      offset   = - scale * min;

      for(i = r0; i <= r1; i++)
	{
	  for(j = c0; j <= c1; j++)
	    {
	      _R[i][j] = (unsigned char)(scale * _A[i][j] + offset);
	    }
	}

  

      DeleteIfTemporary(A);
  
  
  
      return ( (*R).temporary() );
    }


  //! Template for unsigned char rescale [0, 255]
  template<class T> array<unsigned char> & LogRescale(array<T> & A)
    {
      int c0 = A.first_col_index();
      int c1 = A.last_col_index();
      int r0 = A.first_row_index();
      int r1 = A.last_row_index();

      array<unsigned char>* R = new array<unsigned char>(r0,r1,c0,c1);
      unsigned char **     _R = R->storage();

      T max, min;
      MinMax(A, min, max);
      if (max == min) return (R->temporary());

      T ** _A =  A.storage();

      double offset   = 1 - min;
      double scale    = 255 / std::log(max + offset);

      for (int i = r0; i <= r1; i++)
	for (int j = c0; j <= c1; j++)
	  _R[i][j] = (unsigned char)(scale * std::log(_A[i][j] + offset));

      DeleteIfTemporary(A);
      return (R->temporary());
    }


  //! Template RGB->Intensity
  //! Usage:
  //! Intensity<int>(R,G,B)
  template<class T, class U> array<T> & Intensity(array<U> & R, 
						  array<U> & G, 
						  array<U> & B)
    {
  
      int i,j;
  
      const int first_row_idx = R.first_row_index();
      const int last_row_idx  = R.last_row_index();
      const int first_col_idx = R.first_col_index();
      const int last_col_idx  = R.last_col_index();
  
  
      array<T> *   I   = new array<T>(first_row_idx,last_row_idx,first_col_idx,last_col_idx); 
      T **        _I   = (*I).storage();
  
      U ** _R             = R.storage();
      U ** _G             = G.storage();
      U ** _B             = B.storage();
  
      for(i = first_row_idx; i <= last_row_idx; i++)
	{
	  for(j = first_col_idx; j <= last_col_idx; j++)
	    {
	      _I[i][j] = (T) ( ( _R[i][j] + _G[i][j] + _B[i][j] ) / 3.0 );
	    }
	}
  


      DeleteIfTemporary(R);
      DeleteIfTemporary(G);
      DeleteIfTemporary(B);
        
      return (  (*I).temporary() );
    }

  //! Template RGB->Intensity
  //! Older version
  template<class T> array<T> & Intensity(array<unsigned char> & R, 
					 array<unsigned char> & G, 
					 array<unsigned char> & B, 
					 array<T> & dummy_forcast)
    {
  
      int i,j;
  

      const int first_row_idx = R.first_row_index();
      const int last_row_idx  = R.last_row_index();
      const int first_col_idx = R.first_col_index();
      const int last_col_idx  = R.last_col_index();
  
      array<T> *   I   = new array<T>(first_row_idx,last_row_idx,first_col_idx,last_col_idx); 
      T **        _I   = (*I).storage();
  
      unsigned char ** _R             = R.storage();
      unsigned char ** _G             = G.storage();
      unsigned char ** _B             = B.storage();
  
      for(i = first_row_idx; i <= last_row_idx; i++)
	{
	  for(j = first_col_idx; j <= last_col_idx; j++)
	    {
	      _I[i][j] = (T) ( ( _R[i][j] + _G[i][j] + _B[i][j] ) / 3.0 );
	    }
	}
  


      DeleteIfTemporary(R);
      DeleteIfTemporary(G);
      DeleteIfTemporary(B);
      DeleteIfTemporary(dummy_forcast);
  
      return (  (*I).temporary() );
  
    }


  template <class T, class C> int array2buffer(array<T>& A, C* buffer)
    {
      T** __A = A.storage();
      T* _A;

      int y0 = A.first_row_index();
      int x0 = A.first_col_index();
      int y1 = A.last_row_index();
      int x1 = A.last_col_index();

      for (int y = y0 ; y <= y1; y++)
	{
	  _A = __A[y] + x0;
	  for(int x = x0; x <= x1; x++) 
	    *buffer++ = (C) *_A++;
	}

      return (A.cols() * A.rows()); 
    }


  template <class T, class C> int buffer2array(const C* buffer, array<T>& A)
    {
      T** __A = A.storage();
      T* _A;

      int y0 = A.first_row_index();
      int x0 = A.first_col_index();
      int y1 = A.last_row_index();
      int x1 = A.last_col_index();

      for (int y = y0 ; y <= y1; y++)
	{
	  _A = __A[y] + x0;
	  for(int x = x0; x <= x1; x++) 
	    *_A++ = (T)*buffer++;
	}

      return (A.cols() * A.rows()); 
    }


  template <class T, class C> int rgb2buffer(array<T>& R, array<T>& G, array<T>& B, C* buffer)
    {
      CheckArraysConsistency(R, G);
      CheckArraysConsistency(R, B);

      T** __R = R.storage();
      T* _R;
      T** __G = G.storage();
      T* _G;
      T** __B = B.storage();
      T* _B;


      int y0 = R.first_row_index();
      int x0 = R.first_col_index();
      int y1 = R.last_row_index();
      int x1 = R.last_col_index();

      for (int y = y0 ; y <= y1; y++)
	{
	  _R = __R[y] + x0;
	  _G = __G[y] + x0;
	  _B = __B[y] + x0;
	  for(int x = x0; x <= x1; x++)
	    { 
	      *buffer++ = (C) *_R++;
	      *buffer++ = (C) *_G++;
	      *buffer++ = (C) *_B++;
	    }
	}

      return (3 * R.cols() * R.rows()); 
    }


  template <class T, class C> int buffer2rgb(const C* buffer, array<T>& R, array<T>& G, array<T>& B)
    {
      CheckArraysConsistency(R, G);
      CheckArraysConsistency(R, B);

      T** __R = R.storage();
      T* _R;
      T** __G = G.storage();
      T* _G;
      T** __B = B.storage();
      T* _B;


      int y0 = R.first_row_index();
      int x0 = R.first_col_index();
      int y1 = R.last_row_index();
      int x1 = R.last_col_index();

      for (int y = y0 ; y <= y1; y++)
	{
	  _R = __R[y] + x0;
	  _G = __G[y] + x0;
	  _B = __B[y] + x0;
	  for(int x = x0; x <= x1; x++)
	    { 
	      *_R++ = (T) *buffer++;
	      *_G++ = (T) *buffer++;
	      *_B++ = (T) *buffer++;
	    }
	}

      return (3 * R.cols() * R.rows()); 
    }

  template<class T, class C> int rgba2buffer(array<T>& R, 
					     array<T>& G,
					     array<T>& B,
					     array<T>& A,
					     C * buffer)
  {
    
    CheckArraysConsistency(R, G);
    CheckArraysConsistency(R, B);
    CheckArraysConsistency(R, A);

    int r,c;
    T ** __R = R.storage();
    T ** __G = G.storage();
    T ** __B = B.storage();
    T ** __A = A.storage();

    T * _R;
    T * _G;
    T * _B;
    T * _A;

    int r0 = R.first_row_index();
    int c0 = R.first_col_index();
    int r1 = R.last_row_index();
    int c1 = R.last_col_index();
    
    for (r = r0 ; r <= r1; ++r)
	{
	  _R = __R[r] + c0;
	  _G = __G[r] + c0;
	  _B = __B[r] + c0;
	  _A = __A[r] + c0;

	  for(c = c0; c <= c1; ++c)
	    { 
	      *buffer++ = (C) *_R++;
	      *buffer++ = (C) *_G++;
	      *buffer++ = (C) *_B++;
	      *buffer++ = (C) *_A++;
	    }
	}

    return (4 * R.rows() * R.cols());
  }

  template<class T, class C> int buffer2rgba(C * buffer,
					     array<T>& R, 
					     array<T>& G,
					     array<T>& B,
					     array<T>& A)
    {
      CheckArraysConsistency(R, G);
      CheckArraysConsistency(R, B);
      CheckArraysConsistency(R, A);

      int r,c;
      T ** __R = R.storage();
      T ** __G = G.storage();
      T ** __B = B.storage();
      T ** __A = A.storage();

      T * _R;
      T * _G;
      T * _B;
      T * _A;

      int r0 = R.first_row_index();
      int c0 = R.first_col_index();
      int r1 = R.last_row_index();
      int c1 = R.last_col_index();
    
      for (r = r0 ; r <= r1; ++r)
	{
	  _R = __R[r] + c0;
	  _G = __G[r] + c0;
	  _B = __B[r] + c0;
	  _A = __A[r] + c0;

	  for(c = c0; c <= c1; ++c)
	    { 
	      *_R++ = (T) *buffer++;
	      *_G++ = (T) *buffer++;
	      *_B++ = (T) *buffer++;
	      *_A++ = (T) *buffer++;
	    }
	}

      return (4 * R.rows() * R.cols());
    }
					     


  template <class T> unsigned int Measure(array<T>& I)
    {
      unsigned int hits = 0;
    
      T ** _I = I.storage();
      int x0  = I.first_col_index();
      int x1  = I.last_col_index();
      int y0  = I.first_row_index();
      int y1  = I.last_row_index();
   
      for (int y = y0; y <= y1; y++)
	for (int x = x0; x <= x1; x++)
	  if (_I[y][x] > 0)
	    hits++;
   
      DeleteIfTemporary(I);
     
      return hits;
    }


  template <class T> inline int IsInImage(int r, int c, array<T> & I)
    {
      if( r >= I.first_row_index() &&
	  r <= I.last_row_index() &&
	  c >= I.first_col_index() &&
	  c <= I.last_col_index() )
	return TRUE;
      else
	return FALSE;
    }
  
  //Oij = Iij if Iij in [min,max] else min or max
  template<class T> array<T> & Force_in_range(array<T> & I, T min, T max)
    {
      T ** _I = I.storage();
      int x0  = I.first_col_index();
      int x1  = I.last_col_index();
      int y0  = I.first_row_index();
      int y1  = I.last_row_index(); 
      array<T> * O = new array<T>(y0,y1,x0,x1);
      T ** _O = O->storage();

      for (int y = y0; y <= y1; y++)
	{
	  for (int x = x0; x <= x1; x++)
	    {
	      _O[y][x] =  RANGE(_I[y][x],min,max);
	    }
	}
      DeleteIfTemporary(I);
      return O->temporary();
    }

  template<class T, class U> bool AreConsistent(array<T> &A, array<U> &B)
    {
      if((A.first_row_index() != B.first_row_index())	||
	 (A.first_col_index() != B.first_col_index())	||
	 (A.rows()		   != B.rows())			||
	 (A.cols()		   != B.cols()))
	{
	  return false;
	}
      return true;
    }
}














#endif
