#
# Implementation of anisotropic (gmap) heat diffusion equation 
# for smoothing with a space variant sigma.
# It has been implemented mainly to support 'advanced' smoothing
# required by non standard sampling (e.g. in the construction of log-polar
# maps)
#
#

#
# This is a sample map that can be used to modulate
# the effect of anisotropic diffusion: it provides
# a gradient along the horizontal direction alone
# going from 0 at the left side to 1 at the right
# side
#

tm.sampleGradientMap <- function(img)
  {
    data <- img@data;
    gmap <- data * 0;
    
    rows <- dim(data)[1]
    cols <- dim(data)[2]
    
    for(x in 1:cols) {
      f <- (x / cols);
      for(y in 1:rows) {
        gmap[y,x] <- f;
      }
    }

    return(gmap);
  }

#
# In order to get a more detailed sigma resolution we may start with
# fh smaller than 0.25: it will then be double at each iteration as long
# as it is smaller than 0.25 (the maximum allowed by stability analysis)
#
# The image itself must be (?) within [0,1].
#
# The algorithm is from
#
# Bertalmio M, Fort P, and Sanchez-Crespo D
# Real-time, Accurate Depth of Field using Anisotropic Diffusion and
# Programmable Graphics Cards
# Proc. of the 2nd International Symposium on 3D Data Processing,
# Visualization, and Transmission (3DPVT'04)
# pp 767-773
#

tm.anisotropicDiffusion <- function(img,          # an animage
                                    smap,         # an array with the circle of confusion
                                    maxSigma = 20,# no more smoothing than ...
                                    fh       = 0.25)
  {
    #
    if((min(smap) < 0) || (max(smap) > 1)) {
      stop("smoothing map values must be within [0,1]")
    }
    #
    # We may work directly on the image ROI
    #
    data <- img@data;
    #
    # and we will need buffering during operation
    #
    I    <- data;
    Iu   <- I * 0;
    Iv   <- Iu;
    Ig   <- Iu;
    #
    # The number of iterations is linearly related to the squared
    # standard deviation 
    #
    g    <- smap;
    #
    rows <- dim(data)[1]
    cols <- dim(data)[2]
    #
    # We adopt a progressive fh factor, and we compute the equivalent
    # sigma(^2) by accumulation of successive contributions
    # 
    s2   <- 0;
    while(sqrt(s2) < maxSigma)
      {
        #
        # Update the anisotropy map.
        #
        # We first map current iteration number into
        # its equivalent standard deviation squared
        #
        s2 <- s2 + 2 * sqrt(fh * 8);
        message(sprintf("Current sigma is %5.2f", sqrt(s2)));
        #
        # Compute the divergence
        #
        for(y in 1:(rows-1)) {
          for(x in 1:cols) {
            Iu[y,x] = I[y+1,x] - I[y,x];
          }
        }
        #
        for(x in 1:cols) {
          Iu[rows,x] = I[rows,x] - I[rows-1,x];
        }
        #
        for(y in 1:rows) {
          for(x in 1:(cols-1)) {
            Iv[y,x] = I[y,x+1] - I[y,x];
          }
        }
        #
        for(y in 1:rows) {
          Iv[y,cols] = I[y,cols] - I[y,cols-1];
        }
        #
        # and modulate it with diffusion
        # blocking information
        #
        Iu <- Iu * g;
        Iv <- Iv * g;
        #
        # In the following computations we switch forward and
        # backward differences at the outer frame so that we
        # may 'diffuse' on it (otherwise the outer frame should
        # be kept unchanged).
        #
        # Compute the gradient
        #
        for(y in 2:rows) {
          for(x in 1:cols) {
            Ig[y,x] = fh * (Iu[y,x] - Iu[y-1,x]);
          }
        }
        #
        for(x in 1:cols)
          Ig[1,x] = fh * (Iu[2,x] - Iu[1,x])
        #
        for(y in 1:rows) {
          for(x in 2:cols) {
            Ig[y,x] = Ig[y,x] + fh * (Iv[y,x] - Iv[y,x-1]);
          }
        }
        for(y in 1:rows)
          Ig[y,1] = Ig[y,1] + fh * (Iv[y,2] - Iv[y,1]);
        # 
        I <- I + Ig;
        #
        # step up till maximum stability is
        # reached
        #
        if(fh < 0.25) {
          fh <- min(0.25, fh * 2);
        } else {
          fh <- 0.25;
        }
        #
        for (y in 1:rows) {
          for(x in 1:cols) {
            if(I[y,x] > 255) {
              I[y,x] <- 255;
            } else {
              if (I[y,x] < 0) {
                I[y,x] <- 0;
              }
            }
          }
        }
      }
    #    
    res <- new("animage",
               data     = I,
               focus    = img@focus,
               outside  = img@outside,
               storage  = "real",
               document = img@document)
    #
    if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
      res@history  <- updatedHistory
      res@document <- TRUE
    } else {
      res@history  <- call("<undef>")
      res@document <- FALSE
    }
    #
    return(res);
  }

#
# Compute a sequence of blurred images using the heat diffusion
# equation based on a provided sigma map: the final image
# is composed by the pixels with the appropriate smoothing
# from the image sequence
#
# Note that a faster (may be less accurate solution) could
# be via interpolation of a discrete set of smoothed images
# (possibly exploiting iterated application of a single
# Gaussian filter)
#
#
tm.spaceVariantDiffusion <- function(img,            # an animage
                                     smap,           # an array
                                     maxSigma = 20,  # no more than 
                                     fh       = 0.25)
  {
    #
    if((min(img@data) < 0) || (max(img@data) > 1)) {
      stop("image values must be within [0,1]")
    }
    #
    # We may work directly on the image ROI
    #
    data <- img@data;
    #
    # and we will need buffering during operation
    #
    I    <- data;
    If   <- I;
    Iu   <- I * 0;
    Iv   <- Iu;
    Ig   <- Iu;
    #
    # The number of iterations is linearly related to the squared
    # standard deviation 
    #
    G    <- smap * smap;
    #
    rows <- dim(data)[1]
    cols <- dim(data)[2]
    #
    # We adopt a progressive fh factor, and we compute the equivalent
    # sigma(^2) by accumulation of successive contributions
    # 
    s2   <- 0;
    t    <- 1;
    while (sqrt(s2) < maxSigma)
      {
        #
        # Update the anisotropy map.
        #
        # We first map current iteration number into
        # its equivalent standard deviation squared
        #
        s2 <- s2 + 2 * sqrt(fh * 8); # would be s2 <- t*2*sqrt(2))
        message(sprintf("Current sigma is %5.2f", sqrt(s2)));
        #
        # and for all map locations
        #
        for (y in 1:rows) {
          for(x in 1:cols) {
            #
            # we progressively update the final image
            #
            if(s2 <= G[y,x]) {
              If[y,x] <- I[y,x]
            }
          }
        }
        #
        # Compute the divergence
        #
        for(y in 1:(rows-1)) {
          for(x in 1:cols) {
            Iu[y,x] = I[y+1,x] - I[y,x];
          }
        }
        for(x in 1:cols) {
          Iu[rows,x] = I[rows,x] - I[rows-1,x];
        }
        #
        for(y in 1:rows) {
          for(x in 1:(cols-1)) {
            Iv[y,x] = I[y,x+1] - I[y,x];
          }
        }
        #
        for(y in 1:rows) {
          Iv[y,cols] = I[y,cols] - I[y,cols-1];
        }
        #
        # In the following computations we switch forward and
        # backward differences at the outer frame so that we
        # may 'diffuse' on it (otherwise the outer frame should
        # be kept unchanged).
        #
        # Compute the gradient
        #
        for(y in 2:rows) {
          for(x in 1:cols) {
            Ig[y,x] = fh * (Iu[y,x] - Iu[y-1,x]);
          }
        }
        #
        for(x in 1:cols)
          Ig[1,x] = fh * (Iu[2,x] - Iu[1,x])
        #
        for(y in 1:rows) {
          for(x in 2:cols) {
            Ig[y,x] = Ig[y,x] + fh * (Iv[y,x] - Iv[y,x-1]);
          }
        }
        #
        for(y in 1:rows)
          Ig[y,1] = Ig[y,1] + fh * (Iv[y,2] - Iv[y,1]);
        #
        I <- I + Ig;
        #
        # step up till maximum stability is
        # reached
        #
        if(fh < 0.25) {
          fh <- min(0.25, fh * 2);
        } else {
          fh <- 0.25;
        }
        #
        for (y in 1:rows) {
          for(x in 1:cols) {
            if(I[y,x] > 255) {
              I[y,x] <- 255;
            } else {
              if (I[y,x] < 0) {
                I[y,x] <- 0;
              }
            }
          }
        }
        t <- t+1
      }
    #
    res <- new("animage",
               data     = If,
               focus    = img@focus,
               outside  = img@outside,
               storage  = "real",
               document = img@document)
    #
    if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
      res@history  <- updatedHistory
      res@document <- TRUE
    } else {
      res@history  <- call("<undef>")
      res@document <- FALSE
    }
    #
    return(res)
  }
