# 
# The  Bresenham line algorithm  is an  algorithm that  determines which
# points in an n-dimensional raster should be plotted in order to form a
# close approximation to a straight line between two given points. It is
# commonly used  to draw  lines on  a computer screen,  as it  uses only
# integer addition, subtraction  and bit shifting all of  which are very
# cheap operations in standard computer  architectures. It is one of the
# earliest algorithms developed in the field of computer graphics.
#
# http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
#

tm.bresenhamLine <- function(xx1, yy1, xx2, yy2, img, color,
                             operation     = "set",
                             trackFunction = FALSE,
                             imagePoint    = FALSE) {
  x1 <- xx1
  y1 <- yy1
  
  x2 <- xx2
  y2 <- yy2

  if(xx1 <= xx2) {
    x1 <- xx1
    y1 <- yy1

    x2 <- xx2
    y2 <- yy2
  } else {
    x1 <- xx2
    y1 <- yy2

    x2 <- xx1
    y2 <- yy1
  }

  dx <- abs(x2-x1)
  dy <- abs(y2-y1)

  if(y2 < y1) {
    inc_dec <- -1
  } else {
    inc_dec <-  1
  }

  if(dx > dy) {
    two_dy    <- 2*dy
    two_dy_dx <- 2*(dy-dx)
    p         <- 2*dy-dx

    x <- x1
    y <- y1

    if(operation == "+") {
      img[y,x] <- img[y,x] + color
    } else {
      img[y,x] <- color
    }

    if(is.function(trackFunction)) {
      trackFunction(x,y, imagePoint)
    }
    
    while(x<x2) {
      x <- x+1

      if(p<0) {
        p <- p+two_dy
      } else {
        y <- y + inc_dec
        p <- p + two_dy_dx
      }
      
      if(operation == "+") {
        img[y,x] <- img[y,x] + color
      } else {
        img[y,x] <- color
      }

      if(is.function(trackFunction)) {
        trackFunction(x,y, imagePoint)
      }

    }
  } else {
    two_dx    <- 2*dx
    two_dx_dy <- 2*(dx-dy)
    p         <- 2*dx-dy

    x <- x1
    y <- y1

    if(operation == "+") {
      img[y,x] <- img[y,x] + color
    } else {
      img[y,x] <- color
    }

    if(is.function(trackFunction)) {
      trackFunction(x,y,imagePoint)
    }

    while(y != y2) {
      y <- y+inc_dec

      if(p<0) {
        p <- p+two_dx
      } else {
        x <- x + 1
        p <- p + two_dx_dy
      }
      
      if(operation == "+") {
        img[y,x] <- img[y,x] + color
      } else {
        img[y,x] <- color
      }

      if(is.function(trackFunction)) {
        trackFunction(x,y,imagePoint)
      }

    }
  }

  img
}

#
# Generate an image of Gaussian noise with the specified standard
# deviation and congruent to the img supplied as argument
#

gaussianNoiseImage <- function(img, sd = 1) {
  data <- img@data

  rows <- dim(data)[1]
  cols <- dim(data)[2]

  for (x in 1:cols) {
    for(y in 1:rows) {
      data[y,x] <- rnorm(1, sd = sd)
    }
  }
  
  animage(data, storage = img@storage, focus = img@focus)
}

#
# Draw the sinusoid for point (x,y) in Hough parameter space
#

tm.drawParameterCurve <- function(img, houghAccumulator, x, y, rMax, value = 1.0, trackFunction = FALSE) {
  #
  rRes <- (dim(img@data)[1]-1)/2
  tRes <- dim(img@data)[2]
  #
  # Precompute cos and sin
  #
  coses <- cos((0:(tRes-1))*(acos(-1)/tRes))
  sines <- sin((0:(tRes-1))*(acos(-1)/tRes))
  scale <- (rRes/rMax)
  #
  r0    <- as.integer(round((x*coses[1]+y*sines[1])*scale))
  #
  for(t in 2:tRes) {
    #
    r1    <- as.integer(round((x*coses[t]+y*sines[t])*scale))
    #
    houghAccumulator <- tm.bresenhamLine(t-1-1, r0,
                                         t-1,   r1,
                                         houghAccumulator, 
                                         value,
                                         "+",
                                         imagePoint = c(x,y),
                                         trackFunction = trackFunction)
    #
    r0 <- r1
    #
  }
  #
  houghAccumulator
  #
}

#
# Compute the Hough transform on img searching for lines
# with the given radial and angular resolution
#

tm.houghTransform <- function(img, tRes, rRes, view = FALSE, trackFunction = FALSE) {
  #
  data <- img@data
  rows <- dim(data)[1]
  cols <- dim(data)[2]
  ox   <- img@focus[1]
  oy   <- img@focus[2]
  #
  rMax <- sqrt(rows*rows+cols*cols)/2
  #
  # create the accumulator
  #
  hough <- animage(matrix(data = 0,
                          nrow=2*rRes+1,
                          ncol=tRes),
                   storage = "real", focus = c(0, -rRes))

  if(is.function(trackFunction))
    {
      tm.houghTemplate <<- ia.mult(hough, 0)
    }
  
  #
  # Scan the image 
  #
  for(i in 1:cols) {
    for(j in 1:rows) {
      #
      # and for each point draw the curve in parameter space
      #
      # if(data[j,i] > 0)
      hough <- tm.drawParameterCurve(img, hough, i+ox-1, j+oy-1,
                                     rMax, data[j,i], 
                                     trackFunction = trackFunction)
      #
      if(view)
        plot(pixmapGrey(hough@data))
    }
  }
  #
  hough
  #
}

myTrack <- function(t,r,imagePoint) {
  if((t == theta) && (r == rho)) {
    tm.houghTemplate[imagePoint[2],imagePoint[1]] <<- 1
  }
}

#
# Compute the length of the line corresponding to
# the (r,t) Hough accumulator cell, assuming a square retina
# (with center (0,0) and diagonal rMax*2+1). 
#
tm.lineLength <- function(r, t, rRes, rMax, tRes) {

  r  <- sqrt(2)*(r / rRes)
  t  <- (t * 3.1415927) / tRes
  c  <- cos(t)
  s  <- sin(t)

  ps <- list(FALSE, FALSE)
  np <- 1
  
  x <- (r - s) / c

  if(abs(x) <= 1) {
    ps[[np]] <- c(x, 1)
    np       <- np+1
  }

  x <- (r + s) / c

  if(abs(x) <= 1) {
    ps[[np]] <- c(x, -1)
    np       <- np+1
  }

  y <- (r - c) / s

  if((abs(y) <= 1) && (np <= 2)){
    ps[[np]] <- c(1, y)
    np       <- np+1
  }

  y <- (r + c) / s

  if((abs(y) <= 1) && (np <= 2))  {
    ps[[np]] <- c(-1, y)
    np       <- np+1
  }
  
  return((rMax/sqrt(2))*sqrt(sum((ps[[1]]-ps[[2]])*(ps[[1]]-ps[[2]]))))
}

#
# Given a Hough accumulator (as produced by tm.houghTransform) generate
# the corresponding reference accumulator containing for each point
# in parameter space the maximal length of the corresponding line assuming
# a square retina whose diagonal is rMax*2+1
# 
tm.houghReference <- function(hough,  rMax) {
  #
  rows <- dim(hough@data)[1]
  cols <- dim(hough@data)[2]
  #
  rRes <- (rows-1)/2
  tRes <- cols
  #
  data <- matrix(data = 0, nrow=rows,ncol=cols)
  #
  for (x in 1:cols) {
    for(y in 1:rows) {
      data[y,x] <- tm.lineLength(y-rRes-1, x-1, rRes, rMax, tRes)
    }
  }
  #
  animage(data, storage = "real", focus = hough@focus)
  #
}

#
# Compute the maximum a posteriori probability exploiting the
# Hough transform.
#
tm.mapHoughTransform <- function(img, tRes, rRes, noiseStdDev, view = FALSE) {
  #
  hough    <- tm.houghTransform(img, tRes, rRes, view = view)
  #
  rows <- dim(img@data)[1]
  cols <- dim(img@data)[2]
  rMax <- sqrt(rows*rows+cols*cols)/2
  #
  houghRef <- tm.houghReference(hough, rMax)
  #
  data    <- hough@data
  dataRef <- houghRef@data
  rows    <- dim(data)[1]
  cols    <- dim(data)[2]
  #
  for(i in 1:cols) {
    for(j in 1:rows) {
      data[j,i] <- exp((1.0/(noiseStdDev*noiseStdDev))*(data[j,i]-0.5*dataRef[j,i]))
    }
  }
  #
  animage(data,
          storage = "real",
          focus = hough@focus)
  #
}

#
# Hough recap plot ...
#
tm.houghPlot <- function(noisyImg, h, hlr) {
  oldpar <- par(no.readonly = TRUE)

  par(mfrow=c(1,3),oma = c(0,0,1,0), mar=c(1,1,1,0),mgp=c(0,0,0))

  plot(pixmapGrey(ia.integerAnimage(ia.stretchRange(noisyImg))@data),
       main = expression(bold(paste("Noisy image (", sigma, "= 0.4)"))))
  persp(h@data, ylab="theta", xlab="rho",
        main = "Std. Hough Transform")
  persp(hlr@data, ylab="theta", xlab="rho",
        main = "L. ratio Hough Transform")
  
  par(oldpar)
}

# cleanLine <- tm.bresenhamLine(-16L,15L, 16L, 5L, img, 1)
# noisyLine <- ia.add(gaussianNoiseImage(cleanLine, sd=1.0),cleanLine)
#
#
#
# Demo (rows correspond to the radial coordinates)
#
# img      <- animage(matrix(data = 0, nrow=33,ncol=33), storage = "real", focus = c(-16L,-16L))
# noisyImg <- tm.addNoise(ia.mult(img, 0), noiseType="normal", scale=0.4)
# h        <- tm.houghTransform(noisyImg, 33, 16, view = FALSE)
# hmap     <- tm.mapHoughTransform(noisyImg, 33, 16, 0.4, view = FALSE)
# tm.houghPlot(noisyImg, h, hmap)
#

