# 
# The function creates the equivalent of a 'raw' version of the
# provided image mimicking a Bayer arrangement of the RGB sensors
#
# If the upper left corner of the image is at (1,1) we get the following
# pattern:
#
#   RG
#   GB
#

tm.bayerizeImage <- function(rgbImage)
  {
    #
    red    <- rgbImage[[1]]@data
    green  <- rgbImage[[2]]@data
    blue   <- rgbImage[[3]]@data
    #
    bayer  <- green
    #
    width  <- dim(red)[2]
    height <- dim(red)[1]
    #
    for (y in 1:height)  {
      for (x in 1:width) {
        pixelType <- ((x-1) %% 2) + ((y-1) %% 2) * 2
        if(pixelType == 0) {         # red
          bayer[y,x] <- red[y,x]
        } else if (pixelType == 1) { # green on RG row
        } else if (pixelType == 2) { # green on GB row
        } else if (pixelType == 3) { # blue
          bayer[y,x] <- blue[y,x]
        }
      }
    }
    #
    res <- new("animage",
               data     = bayer,
               pmode    = rgbImage[[1]]@pmode,
               focus    = rgbImage[[1]]@focus,
               outside  = rgbImage[[1]]@outside,
               storage  = rgbImage[[1]]@storage,
               document = rgbImage[[1]]@document)
    #
    return(res)
    #
  }

#
# Debayerize a 'raw' image using the method proposed by:
#
# Hamilton, Jr., John F. (Rochester, NY, US)
# Adams, Jr., James E. (Rochester, NY, US)
# Adaptive color plane interpolation in single sensor color electronic camera
# U.S.Patent 5,629,734
# May 1997
#
# Assignee Eastman Kodak Company (Rochester, NY)
#
tm.debayerizeImage <- function(img)
  {
    #
    # We need a 2-pixel frame around the original image
    #
    m   <- ia.mask(img)
    img <- ia.get(img,
                  animask(img@focus[1]-2,
                          img@focus[2]-2,
                          dim(img@data)[2]+4,
                          dim(img@data)[1]+4))
    #
    R <- B <- G <- S <- img@data
    #
    width  <- dim(R)[2]
    height <- dim(R)[1]
    #
    # we first need to complete the green channel
    #
    for (y in 3:(height-2))  {
      for (x in 3:(width-2)) {
        pixelType <- ((x-1) %% 2) + ((y-1) %% 2) * 2
        #
        if ((pixelType == 3) ||         # if blue
            (pixelType == 0)) {         # or red
          #
          # we need to compute green
          #
          dh <- (abs(S[y,x-1] - S[y,x+1]) +
                 abs(2*S[y,x] - S[y,x-2] - S[y,x+2]))
          dv <- (abs(S[y-1,x] - S[y+1,x]) +
                 abs(2*S[y,x] - S[y-2,x] - S[y+2,x]))
          #
          if (dh < dv) {
            G[y,x] <- ((S[y,x-1] + S[y,x+1]) / 2 +
                       (2*S[y,x] - S[y,x-2] - S[y,x+2]) / 4)
          } else if (dh > dv) {
            G[y,x] <- ((S[y-1,x] + S[y+1,x]) / 2 +
                       (2*S[y,x] - S[y-2,x] - S[y+2,x]) / 4)
          } else {
            G[y,x] <- ((S[y,x-1] + S[y,x+1] + S[y-1,x] + S[y+1,x]) / 4 +
                       (4*S[y,x] - S[y-2,x] - S[y+2,x] - S[y,x-2] - S[y,x+2]) / 8)
          }
        }       
      }
    }
    #
    for (y in 3:(height-2))  {
      for (x in 3:(width-2)) {
        pixelType <- ((x-1) %% 2) + ((y-1) %% 2) * 2
        #
        if (pixelType == 1) {
          #
          # green on RG row
          # compute red  (nearest neighbors on same row)
          #
          R[y,x] <- ((S[y,x-1]+S[y,x+1]) / 2 +
                     (2*G[y,x] - G[y,x-1] - G[y,x+1]) / 4)
          #
          # compute blue (nearest neighbors on same column)
          #
          B[y,x] <- ((S[y-1,x]+S[y+1,x]) / 2 +
                     (2*G[y,x] - G[y-1,x] - G[y+1,x]) / 4)
          
        } else if (pixelType == 2) {
          #
          # green on GB row
          # compute red  (nearest neighbors on same column)
          #
          R[y,x] <- ((S[y-1,x]+S[y+1,x]) / 2 +
                     (2*G[y,x] - G[y-1,x] - G[y+1,x]) / 4)
          #
          # compute blue (nearest neighbors on same row)
          #
          B[y,x] <- ((S[y,x-1]+S[y,x+1]) / 2 +
                     (2*G[y,x] - G[y,x-1] - G[y,x+1]) / 4)
        } else if ((pixelType == 3) ||  # blue
                   (pixelType == 0)) {  # red
          #
          # compute gradients
          #
          dn <- (abs(S[y-1,x-1] - S[y+1,x+1]) +
                 abs(2*G[y,x] - G[y-1,x-1] - G[y+1,x+1]))
          dp <- (abs(S[y-1,x+1] - S[y+1,x-1]) +
                 abs(2*G[y,x] - G[y-1,x+1] - G[y+1,x-1]))
          #
          if(pixelType == 0) {
            R[y,x] <- S[y,x]
            if(dn > dp) {
              B[y,x] <- ((S[y-1,x+1] + S[y+1,x-1]) / 2 +
                         (2*G[y,x] - G[y-1,x+1] - G[y+1,x-1]) / 2)
            } else {
              B[y,x] <- ((S[y-1,x+1] + S[y+1,x-1] + S[y-1,x-1] + S[y+1,x+1]) / 4 +
                         (4*G[y,x]-G[y-1,x+1]-G[y+1,x-1]-G[y-1,x-1]-G[y+1,x+1])/4)
            }
          } else {
            B[y,x] <- S[y,x]
            if(dn > dp) {
              R[y,x] <- ((S[y-1,x+1] + S[y+1,x-1]) / 2 +
                         (2*G[y,x]-G[y-1,x+1]-G[y+1,x-1]) / 2)
            } else {
              R[y,x] <- ((S[y-1,x+1] + S[y+1,x-1] + S[y-1,x-1] + S[y+1,x+1]) / 4 +
                         (4*G[y,x]-G[y-1,x+1]-G[y+1,x-1]-G[y-1,x-1]-G[y+1,x+1])/4)
            }
          }
        }        
      }
    }
    #
    # transform the arrays into animages
    #
    R <- new("animage",
             data     = R,
             pmode    = img@pmode,
             focus    = img@focus-2L,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)
    #
    G <- new("animage",
             data     = G,
             pmode    = img@pmode,
             focus    = img@focus-2L,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)
    #
    B <- new("animage",
             data     = B,
             pmode    = img@pmode,
             focus    = img@focus-2L,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)
    #
    res <- list(ia.get(R, m),
                ia.get(G, m),
                ia.get(B, m))
    #
    return(res)
    #
  }
