#
# Computes the kernel corresponding to the Airy disc, the results of
# the Fraunhofer diffraction from a circular aperture.
#
# The default pixelSize corresponds to  a Canon 20D
#
# Another example (with data from the Nikon Coolpix 8800) is
#
# tm.diffractionKernel(focalLength=8.9e-3,
#                      fStop=8,
#                      pixelSize=(8.8/3264)/1000)
#
# The focal length for the Nikon Coolpix 8800 varies in the range
# 8.9 - 89 mm (equivalent to 35-350 mm for the 35mm format)
#
# Nikon D70s:
#
# tm.diffractionKernel(focalLength=50e-3,
#                      fStop=8,
#                      pixelSize=(23.7/3008)/1000)
#
tm.diffractionKernel <- function(focalLength   = 50e-3,
                                 fStop         = 11,
                                 wavelength    = 550e-9,
                                 pixelSize     = 6.4e-6,
                                 superSampling = 1)
{
  #
  d <- focalLength / fStop
  a <- d / 2
  k <- (2*pi) / wavelength
  #
  l <- 6 * superSampling
  #
  kernel <- array(c(0), dim = c(2*l+1, 2*l+1))
  #
  for (y in 0:l) {
    for (x in 0:l) {
      r <- 1e-9 + sqrt(x*x + y*y) * (pixelSize / superSampling) / focalLength
      kernel[l+1+y,l+1+x] <- (2 * besselJ(r*a*k, 1) / (a*k*r))**2
      kernel[l+1-y,l+1+x] <- kernel[l+1+y,l+1+x]
      kernel[l+1-y,l+1-x] <- kernel[l+1+y,l+1+x]
      kernel[l+1+y,l+1-x] <- kernel[l+1+y,l+1+x]
    }
  }
  #
  kernel / sum(kernel)
  #
}
