#
# \litcode{Hierachical template matching}
#
# This function illustrates how we can achieve a significant speed up
# in correlation matching by using a multiresolution strategy: we
#
tm.hierarchicalMatching <- function(I, T, levels = 3, thr = 90) {
  #
  # We keep track of the original position of the image region
  # of interest:
  #
  ofocus  <- I@focus
  #
  # and consider it (temporarily) as our new coordinate origin
  #
  I@focus <- c(0L, 0L)
  T@focus <- c(0L, 0L)
  #
  # The first step is to build the multiresolution
  # representations of the image and of the template by carefully
  # smoothing them at each resolution step before subsampling them:
  #
  pyrI    <- tm.gaussianPyramid(I, levels)
  pyrT    <- tm.gaussianPyramid(T, levels)
  #
  # The first element of the resolution list (the pyramid) is the
  # one at the lowest resolution and we create the corresponding computation
  # filter encompassing all pixels:
  #
  mask    <- as.animage(array(1, dim(pyrI[[1]]@data)))
  #
  # When we map the mask to the next higher resolution level
  # we want to slightly enlarge it to avoid missing the right template position.
  # Mathematical morphology provides an easy way to implement the growing operation
  # by means of a simple structuring element and using the \code{ia.mmBitDilation}
  # function:
  #
  eltA <- animage(array(1L, c(3,3)), focus=c(-1L,-1L))
  #
  # We start from the lowest resolution level,
  #
  for(l in 1:levels) {
    #
    # performing correlation only an selected pixels
    #
    cor  <- ia.correlation(pyrI[[l]], pyrT[[l]], filter=mask)[[1]]
    # 
    # and compute the mask for the next (higher) resolution level:
    #
    mask <- ia.integerAnimage(ia.greater(cor, thr))
    mask <- ia.realAnimage(ia.pzoom(ia.mmBitDilation(mask, eltA), 2L, 2L))
  }
  #
  # finally returning the correlation map, at the correct position
  # in the original coordinate system:
  #
  cor@focus <- ofocus
}
