#
# Histogram format:
#
#  list(hoffset=<integer>, hdata=<1D array>
#
tm.histogram <- function(img,
                         mask     = FALSE,
                         srcRange = c(0,255),
                         dstRange = c(0,255)) {
  #
  # standardize image range
  #
  data <- ia.mapRange(ia.integerAnimage(img),
                      srcRange=srcRange,
                      dstRange = dstRange)@data
  #
  # and compute the histogram
  #
  rows <- dim(data)[1]
  cols <- dim(data)[2]
  #
  histo <- array(0, dim=c(1+dstRange[2]-dstRange[1]))
  off   <- dstRange[1]-1
  if(is.animage(mask) == TRUE) {
    maskData <- ia.integerAnimage(mask)@data
    for (y in 1:rows) {
      for (x in 1:cols) {
        if(maskData[y,x] != 0) {
          b        <- data[y,x]-off
          histo[b] <- histo[b] +1
        }
      }
    }
  } else {
    for (y in 1:rows) {
      for (x in 1:cols) {
        b        <- data[y,x]-off
        histo[b] <- histo[b] +1
      }
    }
  }
  #
  list(hoffset = dstRange[1],
       hdata   = histo)
}

tm.cumulateHistogram <- function(histo) {
  hdata = histo$hdata
  n     = length(hdata)
  
  for(b in 2:n)
    hdata[b] = hdata[b-1]+hdata[b]

  list(hoffset = histo$hoffset,
       hdata   = hdata / hdata[n])
}

tm.ehUniform <- function(f,
                         eqSpecs = list(gMin=0, gMax=255)) {
  gMin = eqSpecs$gMin
  gMax = eqSpecs$gMax

  gMin + f * (gMax - gMin)
}

tm.ehExponential <- function(f,
                             eqSpecs = list(gMin=0, alpha=0.01)) {
  gMin  = eqSpecs$gMin
  alpha = eqSpecs$alpha

  gMin - (log(1-f) / alpha)
}

tm.ehRayleigh <- function(f,
                          eqSpecs = list(gMin=0, alpha=50)) {
  gMin  = eqSpecs$gMin
  alpha = eqSpecs$alpha

  gMin + sqrt(-2*alpha*alpha*log(1-f))
}

tm.ehCubeRoot <- function(f,
                          eqSpecs = list(gMin=0, gMax=256)) {
  gMin  = eqSpecs$gMin
  gMax  = eqSpecs$gMax
  alpha = 1.0 / 3
  
  (gMin**alpha + f * (gMax**alpha - gMin**alpha))**3
}

tm.ehLogarithmic <- function(f,
                             eqSpecs = list(gMin=1, gMax=256)) {
  gMin  = eqSpecs$gMin
  gMax  = eqSpecs$gMax

  gMin * ((gMax / gMin) ** f)
}

tm.ehGaussian <- function(f,
                          eqSpecs = list(gMin=0, gMax=256, mu=128, sd=30)) {
  gMin  = eqSpecs$gMin
  gMax  = eqSpecs$gMax
  mu    = eqSpecs$mu
  sd    = eqSpecs$sd
  
  max(gMin,min(gMax, qnorm(f, mu, sd))) 
}

tm.histogramEqualizeMap <- function(histo,
                                    eqSpecs =  list(type="Gaussian", gMin=0,gMax=256, mu=128, sd=30)) {
  #
  # Compute  cumulative distribution
  #
  histoD    <- histo$hdata
  cumHistoD <- histoD;
  bins      <- length(histoD);
  #
  for (b in 2:bins) {
    cumHistoD[b] <- cumHistoD[b-1]+cumHistoD[b]
  }
  #
  # we make the sum slightly less than 1 (by 1/2 pixel)
  # so that we do not incur into problems with some of
  # the available equalization maps
  #
  cumHistoD <- cumHistoD / (cumHistoD[bins] + 0.5)
  #
  map    <- array(0, dim=c(256))
  #
  # get mapping function
  #
  if(eqSpecs$type == "uniform") {
    mapFun <- tm.ehUniform
  } else if (eqSpecs$type == "Rayleigh") {
    mapFun <- tm.ehRayleigh
  } else if (eqSpecs$type == "cubeRoot") {
    mapFun <- tm.ehCubeRoot
  } else if (eqSpecs$type == "exponential") {
    mapFun <- tm.ehExponential
  } else if (eqSpecs$type == "logarithmic") {
    mapFun <- tm.ehLogarithmic
  } else if (eqSpecs$type == "Gaussian") {
    mapFun <- tm.ehGaussian
  } else if (eqSpecs$type == "target") {
    cumH   = tm.cumulateHistogram(eqSpecs$histo)
    #
    # get rid of duplicate values
    #
    H = cumH$hdata
    n = length(H)
    #
    # find highest/lowest indexes with 0/1 value
    #
    i0 = 1
    i1 = n
    #
    for(i in 1:n) 
      if(H[i] <= 0)
        i0 = i
      else
        break
    #
    for(i in n:1) 
      if(H[i] >= 1.0)
        i1 = i
      else
        break
    #
    i0 = i0-1
    i1 = i1-1
    o  = cumH$hoffset
    y   = c(i0+o)
    x   = c(0.0)
    #
    for(i in i0:i1) {
      if(H[i] != x[[1]]) {
        x = c(H[i], x)
        y = c(i+o, y)
      }
    }

    mapFun <- function(f, data) {splinefun(x, y)(f)}
  }

  for(i in 1:256) {
    map[i] <- mapFun(cumHistoD[i], eqSpecs)
  }

  list(lutOffset=as.integer(histo$hoffset), lutData=as.integer(map))
}

tm.histogramEqualize <- function(img,
                                 eqSpecs = list(type="Gaussian", gMin=0, gMax=256, mu=128, sd=30)) {
  hEqualizingMap <- tm.histogramEqualizeMap(tm.histogram(img), eqSpecs = eqSpecs)
  
  ia.table(img,
           hEqualizingMap$lutData,
           hEqualizingMap$lutOffset)
}
