\name{Log Polar Sensors}

\alias{Log Polar Sensors}
\alias{tm.logPolarRings}
\alias{tm.logPolarCoordinates}
\alias{tm.logPolarSampleCoordinates}
\alias{tm.logPolarSmoothing}
\alias{tm.logPolarMap}
\alias{tm.logPolarMapEquivalent}
\alias{tm.logPolarToCartesian}

\title{Simulation of Log-Polar Sensors}

\description{

  Anti aliased implementation of log-polar mapping.
  
}

\usage{

tm.logPolarRings(R, angularResolution = 72)
tm.logPolarCoordinates(x, y, R, angularResolution = 72)
tm.logPolarSampleCoordinates(x, y, R, angularResolution = 72)
tm.logPolarToCartesian(i, itheta, R, angularResolution = 72)

tm.logPolarSmoothing(img, angularResolution = 72)

tm.logPolarMap(simg, R, angularResolution = 72)
tm.logPolarMapEquivalent(simg, R, angularResolution = 72, outside = 255)

}

\arguments{

  \item{x}{a real representing the horizontal Cartesian image coordinate}
  \item{y}{a real representing the horizontal Cartesian image coordinate}
  \item{i}{a real (integer) representing the log polar ring}
  \item{itheta}{a real (integer) representing the log polar sector}
  
  \item{angularResolution}{a real representing the (constant) angular
	resolution of the sensor}
  \item{R}{a real representing the max radius of the log polar sensor}

  \item{img}{a real \code{animage}}
  \item{simg}{a real \code{animage} representing a properly smoothed
	(anti-aliased) version of the image}

  \item{outside}{a real representing the value of the pixels outside the
  log polar sensors field}
}

\value{
  
  The functions (right) return several kind of values (left):

  \item{integer}{\code{tm.logPolarRings}}
  \item{vector}{\code{tm.logPolarCoordinates}, \code{tm.logPolarCoordinates}}
  \item{animage}{\code{tm.logPolarSmoothing},\code{tm.logPolarMapEquivalent}}

}

\details{

  The functions enable the mapping from a usual cartesian image,
  whose pixels can be addressed by means of rectangular coordinates, to
  a log polar  representation that is characterized by  pixels placed in
  concentric rings with the same number of pixels in each ring. The
  representations if then  forcedly of variant spatial resolution with
  the outer pixels corresponding to larger areas of the original
  cartesian image, and the inner ones providing high resolution,
  foveated details.

  The function \code{tm.logPolarRings} returns the number of rings for a
  log polar sensor of radius \code{R} and given angular resolution. The
  function \code{tm.logPolarCoordinates} maps the cartesian coordinates
  into the couple (ring, sector) white the function
  \code{tm.logPolarSampleCoordinates} returns the cartesian coordinates
  of the point used to describe the image value in the log polar pixel
  associated to the selected cartesian coordinates. As the mapping is
  spatio variant, proper anti aliasing requires smoothing of the image
  eat different levels at different positions, with the smoothing amount
  related to the size of the corresponding image pixel. The function
  \code{tm.logPolarSmoothing} achieves this by means of a smoothing
  function (\code{tm.spaceVariantDiffusion}) based on the heat diffusion
  equation (\code{tm.anisotropicDiffusion}).

  The actual log polar mapping is provided by the function
  \code{tm.logPolarMap} that returns an array whose horizontal
  coordinate corresponds to rings while the vertical one corresponds to
  sectors.

  The cartesian  back projection of the  log polar image  is provided by
\code{tm.logPolarMapEquivalent}  while the  transformation of  log polar
coordinates   to   cartesian   ones   is  provided   by   the   function
\code{tm.logPolarToCartesian}.

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
  \link{Anisotropic Diffusion},
  \code{\link{tm.spaceVariantDiffusion}},
}

\references{

  Rojer, A.S., Schwartz,  E.L. (1990).
  \emph{Design  Consideration  for a  Space-Variant  Visual Sensor  with
Complex-Logarithmic Geometry},
  Proc.  of  the 10th Int.  Conference on Pattern  Recognition, {\bold
    2}:278-285

}

\examples{
#
# Get sample images
#
tiffdir          <- file.path(system.file(package = "TeMa"),
                              "sampleimages/cameraSimulation")
#
i_persp          <- ia.scale(ia.readAnimage(file.path(tiffdir,
                                            "persp_iso12233_large.tif")),
                             maxValue = 255)
#
# Now we want to experiment with log polar mapping
# we get a new smaller image detail to save on computation time
#
i_perspRedDetail <- ia.get(i_persp[[1]], animask(472,272,257,257))
#
# smooth it for an angular resolution of 72 pixels
#
i_perspRedDetail_smoothed <-
  tm.logPolarSmoothing(i_perspRedDetail, angularResolution = 72)
#
# and project/back project to get a feeling of the log polar mapping with
# 3960 pixels instead of 51471
#
i_perspRedDetail_lpe <-
  tm.logPolarMapEquivalent(i_perspRedDetail_smoothed, 128, 72)
#
# and now crate the 'real' log polar map where the horizontal
# coordinate corresponds to rings, while the vertical one
# represents the angular 'sector'
#
i_perspRedDetail_lpm <-
tm.logPolarMap(i_perspRedDetail_smoothed, 128, angularResolution = 72)
#
# and we plot it with a gamma correction of 1.8
#
plot(pixmapGrey((i_perspRedDetail_lpm)**(1.0/1.8)),
     main="Log polar map with k=72")
#
# comparing it to the cartesian back projection
# (with a gamma correction of 1.8)
#
plot(pixmapGrey((i_perspRedDetail_lpe@data)**(1.0/1.8)),
     main="Cartesian equivalent of a log polar map with k=72")
#

}

\keyword{array}



