#  \litcode{On covariance estimation errors}
#
# We want to visualize the impact of errors in the estimation
# of the covariance matrix on $P_D$, the detection probability, 
# at different operating conditions $P_F$, as typical in
# the Neyman-Pearson paradigm.
#
# We refer to a single dimensional case, assuming that the
# distance of  class means is 1: the difficulty of the problem
# is changed by changing the (common) standard deviation $\sigma$ describing
# the two distributions. As easily checked from the results of
# Section~\TM{3.3}, we have that  $\sigma_0 = \sigma^{-1}$. We want to
# compute the impact on $P_D(\sigma_0)$ given that we set the operating condition
# using $P_F(\sigma')$, where $\sigma'$ is our estimate of $\sigma_0$
#
# We first need to define a few functions, corresponding to Equation~\TM:{3.15},
#
tm.Q <- function(x, sd = 1.0) {
  1 - pnorm(x/sd)
}
#
# to Equation~\TM:{3.57},
#
tm.Pf <- function(nu, sd0 = 1.0) {
  tm.Q(nu / sd0 + sd0/2)
}
#
# and to Equation~\TM:{3.58},
#
tm.Pd <- function(nu, sd0 = 1.0) {
  tm.Q(nu / sd0 - sd0/2)
}
#
# By choosing $\sigma = 1/3$, from which $\sigma_0 = 3$,
# we get a reasonable testing case:
#
tm.covarianceImpact <- function(sigma0 = 3,
# We then consider $P_F \in [0.01, 0.3]$ 
                                pfRange = c(0.01, 0.3, 0.01),
# and a moderately large range for $\sigma' = \alpha \sigma_0$
# $\alpha \in [0.8, 1.2]$:
                                pcRange = c(0.8, 1.2, 0.025)) {
  # We generate the sampling sequences:
  ats <- seq(-3*sigma0, 3*sigma0, 0.1)
  pfs <- do.call(seq, as.list(pfRange))
  pcs <- do.call(seq, as.list(pcRange)) * sigma0
  # and determine their lengths
  npfs <- length(pfs)
  npcs <- length(pcs)
  # from which we appropriately size the map:
  ci <- array(0, dim=c(npfs, npcs))
  #
  # We now hypothesize several estimated values $\sigma'$, 
  #
  for(c in 1:npcs) {
    #
    # and for each of them, we build the function \code{nu}
    # $\nu = \nu(x): P_F(\nu) = x$:
    #
    nu  <- splinefun(tm.Pf(ats, sd0=pcs[c]), ats) 
    for(f in 1:npfs) {
      #
      # We can now compute for a selected subset of operating
      # conditions $P_F(\sigma')$ based on our estimated standard
      # deviation, the difference in the miss probability with
      # respect to the correct one $P_D(\sigma_0)$:
      #
      ci[f,c] <- (1-tm.Pd(nu(pfs[f]), pcs[c])) /(1-tm.Pd(nu(pfs[f]), sigma0))-1
    }
  }
  list(pfs, pcs, ci)
}
