#
# We want to generate distortion maps from a given radial description.
# We will need:
#
#                  the width
#                  the height
#
# of the sensor, as the generated maps will have the same size, we then need
# the scaling factor to map actual spline data to the simulated sensor (e.g.
# actual data will be for a 24x36 mm frame, so the scaling factor will be
# 24 / height: for each pixel of coordinates y in our sensor we want to go
# back to r = 24 * y / height in our table.
#
# The same will apply for relative illuminance.
#
# Then we need the distortion (vignetting) data: we will represent them with
# an interpolator.
# In the preferred constructor function we will default it to no distortion 
#

#
# Distagon T* f/2.8 - 21mm Zeiss data
#
distagonDistortion <- splinefun(c(0,  5.00, 10.00, 11.90, 15.00, 18.30, 20.00, 21.60),
                                c(0, -0.41, -1.55, -2.00, -2.45, -2.00, -1.18, -0.11))
distagonVignette   <- splinefun(c(0,  5.00, 10.00, 15.00, 20.00, 21.60),
                                c(1,  0.86,  0.64,  0.41,  0.23,  0.18))
#
# We compute the deformation maps given the radial distortion curve
# Dimensions should be odd: optical center is assumed to be at array center
# We must provide image dimensions and the size of the sensor from
# which the distortion curve is derived (scaling).
#

tm.computeDistortionMaps <- function(width = 301, height = 201,
                                     scaling, curve,
                                     rmax = 21.6)
{
  #
  if(missing(scaling)) {
    #
    # let us assume that distortion data / image dimensions correspond to
    # 135 film (24mm x 36mm)
    #
    scaling = 24 / height;
  }
  #
  halfWidth  <- as.integer((width  - (width  %% 2))/ 2);
  halfHeight <- as.integer((height - (height %% 2))/ 2);
  mapX       <- array(c(0),c(2*halfHeight+1,2*halfWidth+1))
  mapY       <- mapX;
  #
  # For efficiency reasons we will work on array structures
  #
  ox         <- halfWidth  + 1
  oy         <- halfHeight + 1
  #
  # Create the corresponding radial curve for backprojection
  #
  rmin       <- 0
  rseq       <- seq(rmin,rmax,0.1*(rmax-rmin))
  #
  # The factor 0.01 is due to the distortion being express as a %
  # and D = (h'-h)/h where h is the radial distance in the distorted image
  #
  inverseDistortionCurve <- splinefun(0.01*curve(rseq)*rseq+rseq, rseq)
  # 
  # Create the maps exploiting specular symmetry around horizontal and
  # vertical axis to improve speed
  #
  mapX[oy, ox] <- ox
  mapY[oy, ox] <- oy
  #
  for (y in 0:halfHeight) {
    #
    if((y%%20) == 0)
      message(sprintf("Line %5d\n", y), appendLF = FALSE)
    #
    y2 <-  y *  y
    yp <-  y + oy
    yn <- oy -  y
    #
    for (x in 0:halfWidth) {
      if(((x+y) != 0) && (mapX[yp,x+ox] == 0)){
        xp           <- x + ox
        xn           <- ox - x
        r            <- sqrt(y2 + x*x)
        R            <- inverseDistortionCurve(r*scaling) / scaling
        X            <- x * R / r
        Y            <- y * R / r
        #
        mapX[yp,xp]  <-   mapX[yn,xp] <-  X + ox
        mapY[yp,xp]  <-   mapY[yp,xn] <-  Y + oy
        mapX[yp,xn]  <-   mapX[yn,xn] <- -X + ox
        mapY[yn,xp]  <-   mapY[yn,xn] <- -Y + oy
        #
        if((x <= halfHeight) && (y <= halfWidth))
           {
            yp <- oy + x
            yn <- oy - x
            xp <- ox + y
            xn <- ox - y
            #
            mapX[yp,xp] <- mapX[yn,xp] <-  Y + ox
            mapY[yp,xp] <- mapY[yp,xn] <-  X + oy
            mapX[yp,xn] <- mapX[yn,xn] <- -Y + ox
            mapY[yn,xp] <- mapY[yn,xn] <- -X + oy
            #
            yp <-  y + oy
            yn <- oy -  y
           }
       }
    }
  }
  #
  list(mapX, mapY);
  #
}

#
# We compute the vignetting map given the radial vignetting curve
# Dimensions should be odd: optical center is assumed to be at array center
# We must provide image dimensions and the size of the sensor from
# which the distortion curve is derived (scaling).
#

tm.computeVignettingMap <- function(width = 301, height = 201, scaling,
                                    curve)
{
  #
  if(missing(scaling)) {
    #
    # let us assume that distortion data / image dimensions correspond to
    # 135 film (24mm x 36mm)
    #
    scaling = 24 / height;
  }
  #
  halfWidth  <- as.integer((width  - (width  %% 2))/ 2);
  halfHeight <- as.integer((height - (height %% 2))/ 2);
  maxR       <- sqrt(halfHeight*halfHeight+halfWidth*halfWidth);
  mapV       <- array(c(0),c(2*halfHeight+1,2*halfWidth+1))
  ox         <- halfWidth  + 1
  oy         <- halfHeight + 1
  #
  for (y in 0:halfHeight) {
    y2 <-  y *  y
    yp <-  y + oy
    yn <- oy -  y
    #
    for (x in 0:halfWidth) {
      if(((x+y) != 0) && (mapV[yp,x+ox] == 0)){
        xp           <- x + ox
        xn           <- ox - x
        #
        r            <- sqrt(y*y + x*x)
        V            <- curve(r*scaling)
        #
        mapV[yp,xp]   <-   mapV[yn,xp] <-  V
        mapV[yp,xn]   <-   mapV[yn,xn] <-  V
        #
        if((x <= halfHeight) && (y <= halfWidth))
          {
            yp <- oy + x
            yn <- oy - x
            xp <- ox + y
            xn <- ox - y
            #
            mapV[yp,xp]   <-   mapV[yp,xn] <-  V
            mapV[yn,xp]   <-   mapV[yn,xn] <-  V
            #
            yp <-  y + oy
            yn <- oy -  y            
          }
      }
    }
  }
  #
  mapV[oy,ox] <- curve(0)
  #
  mapV
  #
}
                
#
# Lanczos kernel coefficient computation (vectorized)
# A value of support = 2 is appropriate for interpolation,
# as higher values tend to provide ringing (they could be useful
# to preprocess images for print).
#
#

tm.Lanczos <- function(r, support = 2)
{
  #
  if(is.vector(r) && (length(r) > 1)) {
    sapply(r, tm.Lanczos, support)
  } else {
    if(r == 0) {
      return(1)
    } else if(abs(r) < support) {
      return(support*sin(pi*r)*sin(pi*r/support)/(pi*pi*r*r))
    } else {
      return(0)
    }
  }
  #
}

#
# Modify the value of the image, typically simulating the radial
# intensity fall off due to optics effects (and lens design)
#
# img: an animage
# map: an array whose dimensions must be congruent to those of img
#      (typically computed by tm.computeVignettingMap)
#

tm.vignetteImage <- function(img, map) {
  #
  if((dim(img@data)[1] != dim(map)[1]) ||
     (dim(img@data)[2] != dim(map)[2]))
    return(FALSE)
  #  
  res <- new("animage",
             data     = img@data * map,
             pmode    = img@pmode,
             focus    = img@focus,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  #
  return(res)
  #
}

#
# Modify the value of the image, typically simulating the radial
# intensity fall off due to optics effects (and lens design)
#
# img:  an animage
# maps: two arrays whose dimensions must be congruent to those of img
#       (typically computed by tm.computeDistortionMaps)
#

tm.distortImage <- function(img, maps) {
  #
  data <- img@data
  #
  # We should check boundaries
  #
  rows <- dim(data)[1]
  cols <- dim(data)[2]
  #
  mapX <- maps[[1]]
  mapY <- maps[[2]]
  dImg <- array(c(255L), dim(data)) # a white background
  #
  # We use Lanczos interpolation (support = 2 for reduced ringing)
  # in the back mapping step
  #
  for (y in 1:rows) {
    #
    if((y%%20) == 0)
      message(sprintf(" Line %5d\n", y), appendLF = FALSE)
    else
      message(".", appendLF = FALSE)
      
    # 
    for(x in 1:cols) {
      xx <- mapX[y,x]
      yy <- mapY[y,x]
      #
      x0 <- ceiling(xx-2)
      x1 <- floor(xx+2)
      y0 <- ceiling(yy-2)
      y1 <- floor(yy+2)
      #
      if((x0 > -2) && (y0 > -2) &&
         (x1 < cols+2) && (y1 < rows+2)) {
        x0 <- min(cols, max(1, ceiling(xx-2)))
        x1 <- min(cols, max(1, floor(xx+2)))
        y0 <- min(rows, max(1, ceiling(yy-2)))
        y1 <- min(rows, max(1, floor(yy+2)))
        #
        rs <- c()
        vs <- c()
        #
        for(j in y0:y1) {
          for(i in x0:x1) {
            rs <- c(sqrt((xx-i)**2+(yy-j)**2), rs)
            vs <- c(data[j,i], vs)
          }
        }
        #
        cs        <- tm.Lanczos(rs, support = 2)
        v         <- max(0, min(255, sum(vs * cs)  / sum(cs)))
        dImg[y,x] <- as.integer(v+0.5)
      }
    }
  }
  #
  res <- new("animage",
             data     = dImg,
             pmode    = img@pmode,
             focus    = img@focus,
             outside  = img@outside,
             storage  = img@storage,
             document = img@document)
  #
  if(is.call(updatedHistory <- ia.updatedHistory(match.call()))) {
    res@history  <- updatedHistory
    res@document <- TRUE
  } else {
    res@history  <- call("<undef>")
    res@document <- FALSE
  }
  #
  return(res)
  #
}

#
# Computes the kernel corresponding to the Airy disc, the results of
# the Fraunhofer diffraction from a circular aperture.
#
# The default pixelSize corresponds to  a Canon 20D
#
# Another example (with data from the Nikon Coolpix 8800) is
#
# tm.diffractionKernel(focalLength=8.9e-3,
#                      fStop=8,
#                      pixelSize=(8.8/3264)/1000)
#
# The focal length for the Nikon Coolpix 8800 varies in the range
# 8.9 - 89 mm (equivalent to 35-350 mm for the 35mm format)
#
# Nikon D70s:
#
# tm.diffractionKernel(focalLength=50e-3,
#                      fStop=8,
#                      pixelSize=(23.7/3008)/1000)
#

tm.diffractionKernel <- function(focalLength   = 50e-3,
                                 fStop         = 11,
                                 wavelength    = 550e-9,
                                 pixelSize     = 6.4e-6,
                                 superSampling = 1)
{
  #
  d <- focalLength / fStop
  a <- d / 2
  k <- (2*pi) / wavelength
  #
  l <- 6 * superSampling
  #
  kernel <- array(c(0), dim = c(2*l+1, 2*l+1))
  #
  for (y in 0:l) {
    for (x in 0:l) {
      r <- 1e-9 + sqrt(x*x + y*y) * (pixelSize / superSampling) / focalLength
      kernel[l+1+y,l+1+x] <- (2 * besselJ(r*a*k, 1) / (a*k*r))**2
      kernel[l+1-y,l+1+x] <- kernel[l+1+y,l+1+x]
      kernel[l+1-y,l+1-x] <- kernel[l+1+y,l+1+x]
      kernel[l+1+y,l+1-x] <- kernel[l+1+y,l+1+x]
    }
  }
  #
  kernel / sum(kernel)
  #
}

# The default parameters correspond to a Canon 20D digital camera
#
# Canon 20D
#
#  sensor        = CMOS
#  fullwell      = 51400
#  pixel spacing = 6.4e-6
#  resolution    = 3504 x 2336 
#
# For a Nikon D70
#
#  sensor        = CCD
#  fullwell      = 48800
#  pixel spacing = 7.9e-6
#  resolution    = 3008 x 2000
#

## tm.photonNoise <- function(x,
##                            maxValue        =   255,
##                            gammaCorrection =   (1.0 / 2.2),
##                            fullWell        = 51400,
##                            fullWellIso     =   100,
##                            iso             =   200)
##   {
##     #
##     lambdaOrig      <- ((x/maxValue)**(1.0/gammaCorrection)
##                         * fullWell
##                         * (fullWellIso / iso))
##     #
##     lambdaPerturbed <- rpois(1, lambdaOrig)
##     #
##     return(maxValue
##            * ((lambdaPerturbed * (iso / fullWellIso)) / fullWell)**gammaCorrection)
##     #
##   }

