#
# \litcode{Hausdorff matching}
#
# This function illustrates the Hausdorff matching technique:
# the template is moved over each image position and, exploiting the euclidean
# distance transform, the selected Hausdorff partial distance is computed.
#
tm.hausdorffMatching <- function(image, template, distance="average", q = 0.95) {
  #
  # make sure it is binary
  #
  binTemplate <- ia.greater(ia.realAnimage(template),0)
  mask        <- (binTemplate@data)>0
  #
  nq <- sum(binTemplate@data) * q
  #
  # The first step is to get the linear dimensions
  # of the template
  #
  dx <- ia.mask(template)@w
  dy <- ia.mask(template)@h
  #
  # We then determine the extent of the image region
  # of interest over which we need to slide our template
  #
  x0 <- ia.xmin(image)
  y0 <- ia.ymin(image) 
  x1 <- ia.xmax(image) - dx
  y1 <- ia.ymax(image) - dy
  #
  # This allows us to  prepare the storage for the similarity scores
  #
  scores <- animage(array(0,dim=c(y1-y0+1,x1-x0+1)),storage="real",focus=image@focus)
  #
  # We compute the distance transform
  #
  distanceMap <- tm.distanceTransform(ia.table(ia.greater(ia.integerAnimage(image),0L), 1:0, 0L))[[3]]
  #
  # and to  perform the actual computation loop,
  #
  for(y in y0:y1) {
    for(x in x0:x1) {
      #
      # sliding along each row
      # and extracting in turn an image portion
      # matching the size of the template
      #
      w <- ia.get(distanceMap, animask(x, y, dx, dy))
      #
      # We modify the template so that it overlaps
      # the extracted region
      #
      template@focus <- w@focus
      #
      # and compute (and store) the matching score
      #
      if(distance == "average") {
        scores[y,x] <- sum(template@data * w@data)
      } else if (distance == "max") {
        scores[y,x] <- max(w@data[mask])
      } else if (distance == "rank") {
        scores[y,x] <- sort(w@data[mask])[nq]
      }
    }
  }
  #
  # We can finally return the image representing the similarity
  # map
  #
  scores
}
