\name{tm.focusBlur}

\alias{tm.focusBlur}
\alias{tm.getDepth}

\title{Depth of focus as post processing}

\description{

  The generation of high quality depth of focus effects is
  very computationally expensive. Image post processing exploiting depth
  map information is much faster and provides accurate enough results in
  many situations.

}

\usage{
tm.getDepth(filename, depthScale = 1, metric = 10)
tm.focusBlur(channels, depthMap, focalLength,
             fStop, focalDistance, scale = (1201 / 36.0), 
             maxRadius = 30)
}

\arguments{
  \item{filename}{a string identifying a \code{ppm} file}
  \item{depthScale}{a real  scaling factor of the depth map}
  \item{scale}{a real scaling factor mapping the (metric) size of the
	sensor onto image pixels}
  \item{channels}{a list of real \code{animage} objects representing the image}
  \item{depthMap}{a real array with metric depth information (in millimiters)}
  \item{focalLength}{a real representing the focal length of the lens expresse in meters}
  \item{fStop}{a real representing the ratio of lens diameter to the aperture}
  \item{focalDistance}{a real the distance at which is the lens is focused}
  \item{maxRadius}{a real clipping value for the required blurring}
  \item{metric}{a real mapping the units of a depth image to millimiters}
}

\value{
  
  The function \code{tm.getDepth} returns a real array while the
  function \code{tm.focusBlur} returns a list of \code{animage}
  channels.
  
}

\details{

  The implementation of depth of focus effects as a post processing step
  requires the availability of a depth map providing, for each pixel in
  the image, the distance from the camera of the nearest object. These
  images must be produced by the scripts controlling the image rendering
  process in the \code{ppm} format, splitting the depth value (as
  integer) across the red, green, and blue channels, 8 bits per
  channel. The function \code{tm.getDepth} reads such a map and
  transforms it into a real array, dividing the reconstructed integer
  values by \code{depthScale}, a multiplication factor used by the
  renderer when generating the depth maps to increase numeric precision.

  The actual blurring function is \code{tm.focusBlur} based on the
  algorithm proposed by Zhou et al. implements a slower (but more
  accurate) variant without exploiting the approximate separability
  of the filter. The \code{scaling} factor is necessary to correctly map
  the circle of confusion, computed at each pixel using the information
  from the depth map, onto image space.

  \bold{We should change from millimiters to meters for compatibility
	with the usage in Imaging}

}

\author{Roberto Brunelli \email{brunelli@fbk.eu}}

\seealso{
  \code{\link{tm.anisotropicDiffusion}}
}

\references{
  Zhou, T., Chen, JX, Pullen, M. (2007).
  \emph{Accurate Depth of Field Simulation in Real Time},
  COMPUTER GRAPHICS forum {\bold{26}}:15-23 
}

\examples{
#
# Get sample images
#
tiffdir          <- file.path(system.file(package = "TeMa"),
                              "sampleimages/cameraSimulation")
#
i_persp          <- ia.scale(ia.readAnimage(file.path(tiffdir,
                                            "persp_iso12233_large.tif")),
                             maxValue = 255)
i_perspRedDetail <- ia.get(i_persp[[1]], animask(0,0,1201,256))
#
i_perspDepth     <-
  tm.getDepth(file.path(tiffdir, "pp_perspLabel_iso12233_large.ppm"),
              depthScale = 40)
#
i_perspDepthDetail         <- ia.get(i_perspDepth,     animask(0,0,1201,256))
i_perspRedDetail           <- ia.scale(i_perspRedDetail, maxValue = 255)

plot(pixmapGrey(i_perspDepthDetail@data), main="Depth map from camera")

i_perspRedDetail_blurred   <-
tm.focusBlur(list(i_perspRedDetail),
             i_perspDepthDetail,
             21.0,
             8,
             50.0,
             scale     = 1201 / 36, # 1201 pixels over a 36 mm wide sensor
             maxRadius = 30.0) # already quite blurred

plot(pixmapGrey(i_perspRedDetail_blurred[[1]]@data),
     main="Focus blur post processing for a 21mm lens")
}

\keyword{array}



