#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <arrayTools/aniCommon.h>


#ifndef MAX
#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#endif

#ifndef MIN
#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#endif

#ifndef ABS
#define ABS(x)    (((x) > 0)   ? (x) : (-(x)))
#endif


extern "C"  {

  // To each non empty histogram bin we associate the corresponding value
  // of the cumulative distribution (we need this to break ties via 
  // scan order position)
  //
  void rankHistogram(ia::vector<int> &H,
					 ia::vector<int> &R)
  {
	int r = 0;

	for(int i = 0; i < 256; i++)
	  if(H[i] > 0) 
		{
		  r    += H[i];
		  R[i]  = r;
		}
  }
  //
  // Compute the Bhat-Nayar distance of two ranked templates.
  // The first template (reference) is represented with the positions
  // of its sorted pixels, the second is a standard image (2D)
  // where each pixel is assigned its rank.
  //
  // Alpha represents the relative weight of the average distance with
  // respect to the standard (max) one: 
  //
  // alpha = 0 -> max
  // alpha = 1 -> average
  //
  double compareRankedTemplates(ia::array<int> &iA,
								ia::array<int> &B,
								double           alpha)
  {
	int             n = B.rows() * B.cols();
	int             d = 0;
	double          D = 0.0, S = 0.0, aD = 0.0;
	ia::vector<int> H(1,n);

	for(int i = 1; i <= n; i++)
	  {
		int r = B[iA[i][1]][iA[i][0]];

		d    += (r > i) ? 1 : 0;
		d    -= H[i];

		H[r]  = 1;
      
		if(d > D)
		  D = d;

		aD += d;
	  }

	S = 1 - ((1-alpha)*D/(n/2) + alpha*aD/((n*n)/4));

	return(S);
  }
  //
  // Efficient ordinal matching exploiting overlapping regions
  // histogram update
  //
  ia::array<double> &bnCorrelation(ia::array<u_char> &I,
								   ia::array<u_char> &A,
								   double             alpha,
								   int               &rx,
								   int               &ry,
								   double            &mr)
  {
	int               tw = A.cols();
	int               th = A.rows();
	int               n  = tw * th;
	int               x0, x1, y0, y1;
	ia::vector<int>   H0(0,255);  // histogram from beginning of previous line
	ia::vector<int>   Hc(0,255);  // current histogram
	ia::vector<int>   R(0,255);   // ranks for each histogram value
	ia::vector<int>   Rc(0,255);  // ranks for each histogram value
	ia::array<int>    B(0,th-1,0,tw-1);
	ia::array<int>    T(0,th-1,0,tw-1);

	//
	// we must first prepare the template for the comparison 
	// (A -> B)
	//
	// get image dimensions
	//
	A.stat(x0, x1, y0, y1);
	//
	// compute histogram 
	//
	for(int y = y0; y <= y1; y++)
	  for(int x = x0; x <= x1; x++)
		Hc[A[y][x]] += 1;
	//
	// rank it
	//
	rankHistogram(Hc, Rc);
	//
	// and create the ordinal template, breaking ties
	// using scan order
	//
	for(int y = y0, j=0; y <= y1; y++, j++)
	  for(int x = x0, i=0; x <= x1; x++, i++)
		{
		  B[j][i]      = Rc[A[y][x]];
		  Rc[A[y][x]] -= 1;
		}
	//
	// We need to map progressive ranks into
	// the pixels of the reference for efficient
	// comparison
	//
	ia::array<int> iB(1,n,0,1);
	//
	for(int j=0; j < th; j++)
	  for(int i=0; i < tw; i++)
		{
		  iB[B[j][i]][0] = i;
		  iB[B[j][i]][1] = j;
		}
	//
	// and switch to the complete image
	//
	// get image dimensions
	//
	I.stat(x0, x1, y0, y1);
	//
	// compute limits for actual computation
	//
	x1 = x1 - tw + 1;
	y1 = y1 - th + 1;
	//
	// allocate result matrix
	//
	ia::array<double> *BN =	new ia::array<double>(y0, y1, x0, x1);
	//
	// compute first histogram 
	//
	for(int y = y0; y < y0+th; y++)
	  for(int x = x0; x < x0+tw; x++)
		H0[I[y][x]] += 1;
	//
	// and scan the whole image
	//
	double maxS = -1;
	//
	for(int y = y0; y <= y1; y++)
	  {
		if(y > y0) 
		  {
			// we must update H0:
			//
			// we remove previous upper row
			// and add the new entry (lowest)
			//
			for(int i = 0; i < tw; i++)
			  {
				H0[I[y-1][x0+i]]--;
				H0[I[y+th-1][x0+i]]++;
			  }
		  }
		// 
		// and we make it current
		//
		Hc = H0;
		//
		for(int x = x0; x <= x1; x++)
		  {
			if(x > x0) 
			  {
				// we must update Hc
				//
				for(int j = 0; j < th; j++)
				  {
					Hc[I[y+j][x-1]]--;
					Hc[I[y+j][x+tw-1]]++;
				  }
			  }
			//
			// and rank its entries
			//
			rankHistogram(Hc, Rc);
			//
			// compute the sorted window
			//
			for(int yy = y, j=0; yy < y+th; yy++, j++)
			  for(int xx = x, i=0; xx < x+tw; xx++, i++)
				{
				  T[j][i]        = Rc[I[yy][xx]];
				  Rc[I[yy][xx]] -= 1;
				}
			//
			// and proceed to match it to reference
			//
			(*BN)[y][x] = compareRankedTemplates(iB, T, alpha);

			if((*BN)[y][x] > maxS)
			  {
				maxS = (*BN)[y][x];
				mr   = maxS;
				rx   = x;
				ry   = y;
			  }
		  }
	  }

	return((*BN).temporary());
  }

  // Wrapper for bnCorrelation
  //
  SEXP tmBhatNayarCorrelation(SEXP animage_I, // image where the template is searched for (INTEGER)
					          SEXP animage_T, // templated to be located                  (INTEGER)
					          SEXP r_alpha)   // max/ave distance mix factor              (REAL)
  {
	SEXP ani_r;
	double alpha = REAL(r_alpha)[0];


	ia::array<u_char>  I  = Cast(r2ia_ii(animage_I), I);
	ia::array<u_char>  T  = Cast(r2ia_ii(animage_T), T);
	ia::array<double>  R;

	int    rx;
	int    ry;
	double mr;

	R = bnCorrelation(I,
					  T,
					  alpha,
					  rx,
					  ry,
					  mr);

	PROTECT(ani_r = NEW_LIST(4));

	SET_ELEMENT(ani_r, 0, ia2r_dr(R));
	SET_ELEMENT(ani_r, 1, ScalarInteger(rx));
	SET_ELEMENT(ani_r, 2, ScalarInteger(ry));
	SET_ELEMENT(ani_r, 3, ScalarReal(mr));

	UNPROTECT(1);

	return(ani_r);
  }
}
